#!/usr/bin/env python
# coding=utf-8
'''
Author: JiangJi
Email: johnjim0816@gmail.com
Date: 2023-12-24 19:13:11
LastEditor: JiangJi
LastEditTime: 2024-01-13 16:03:27
Discription: 
'''
import numpy as np
from joyrl.algos.base.data_handler import BaseDataHandler


class DataHandler(BaseDataHandler):
    def __init__(self,cfg) -> None:
        self.cfg = cfg
        self.buffer = []
        self.data_after_train = {}
    def add_exps(self, exps):
        ''' add transition to buffer
        '''
        self.buffer.append(exps)

    def sample_training_data(self):
        ''' sample training data from buffer
        '''
        if len(self.buffer) == 0:
            return None
        exp = self.buffer.pop()[0]
        return self._handle_exps_before_train(exp)
    def _handle_exps_before_train(self, exp, **kwargs):
        ''' convert exps to training data
        '''
        state = np.array(exp.state)
        action = np.array(exp.action)
        reward = np.array(exp.reward)
        next_state = np.array(exp.next_state)
        done = np.array(exp.done)
        data = {'state': state, 'action': action, 'reward': reward, 'next_state': next_state, 'done': done}
        return data
    def handle_exps_after_train(self):
        ''' handle exps after train
        '''
        pass
    


#!/usr/bin/env python
# coding=utf-8
'''
Author: JiangJi
Email: johnjim0816@gmail.com
Date: 2023-12-02 15:02:30
LastEditor: JiangJi
LastEditTime: 2024-06-14 09:40:42
Discription: 
'''
import numpy as np
from joyrl.framework.message import Msg, MsgType
from joyrl.framework.config import MergedConfig
from joyrl.framework.base import Moduler
from joyrl.framework.utils import exec_method

class Learner(Moduler):
    ''' learner
    '''
    def __init__(self, cfg : MergedConfig, **kwargs) -> None:
        super().__init__(cfg, **kwargs)
        self.id = kwargs.get('id', 0)
        self.policy = kwargs.get('policy', None)
        self.policy_mgr = kwargs.get('policy_mgr', None)
        self.collector = kwargs.get('collector', None)
        self.data_handler = kwargs.get('data_handler', None)
        self.tracker = kwargs.get('tracker', None)
        self.recorder = kwargs.get('recorder', None)
        self.training_data_que = kwargs.get('training_data_que', None)
        self._init_update_steps()
        self._load_model_meta()
        exec_method(self.logger, 'info', 'remote', f"[Learner.__init__] Start learner {self.id}!")

    def _init_update_steps(self):
        if not self.cfg.is_learner_async:
            self.n_update_steps = 1
        else:
            self.n_update_steps = float('inf')

    def _load_model_meta(self):
        if self.cfg.load_checkpoint and self.cfg.restore_model_meta:
            model_meta = self.cfg.model_meta.get(self.name, {})
            self.policy.load_model_meta(model_meta)
            
    def run(self):
        run_step = 0
        while True:
            training_data = None
            if self.cfg.is_learner_async:
                training_data = exec_method(self.training_data_que, 'pop', 'get')
                # if not self.training_data_que.empty():
                #     training_data = self.training_data_que.get(block=True, timeout=0.1)
            else:
                training_data = exec_method(self.collector, 'pub_msg', 'get', Msg(type = MsgType.COLLECTOR_GET_TRAINING_DATA))
            if training_data is not None:
                self.policy.learn(**training_data)
                exec_method(self.tracker, 'pub_msg', 'remote', Msg(type = MsgType.TRACKER_INCREASE_UPDATE_STEP))
                global_update_step = exec_method(self.tracker, 'pub_msg', 'get', Msg(type = MsgType.TRACKER_GET_UPDATE_STEP))
                # put updated model params to policy_mgr
                model_params = self.policy.get_model_params()
                exec_method(self.policy_mgr, 'pub_msg', 'remote', Msg(type = MsgType.POLICY_MGR_PUT_MODEL_PARAMS, data = (global_update_step, model_params)))
                if global_update_step % self.cfg.policy_summary_fre == 0:
                    avg_model_step_training_data = np.mean(training_data['model_steps'])
                    avg_train_lag = global_update_step - avg_model_step_training_data
                    exec_method(self.logger, 'info', 'remote', f"[Learner.run] update_step: {global_update_step}, avg_train_lag: {avg_train_lag:.2f}")
                    policy_summary = self.policy.get_summary()
                    policy_summary.update({'avg_train_lag': avg_train_lag})
                    summary_data = [(global_update_step,self.policy.get_summary())]
                    exec_method(self.recorder, 'pub_msg', 'remote', Msg(type = MsgType.RECORDER_PUT_SUMMARY, data = summary_data))
                    model_meta = self.policy.get_model_meta()
                    exec_method(self.policy_mgr, 'pub_msg', 'remote', Msg(type = MsgType.POLICY_MGR_PUT_MODEL_META, data = (self.name, model_meta)))
            run_step += 1
            if run_step >= self.n_update_steps:
                break
    


#!/usr/bin/env python
# coding=utf-8
'''
Author: JiangJi
Email: johnjim0816@gmail.com
Date: 2024-06-03 13:37:11
LastEditor: JiangJi
LastEditTime: 2024-06-03 13:37:11
Discription: 
'''

class AlgoConfig(object):
    def __init__(self):
        self.independ_actor = True # whether to use independent actor
        # whether actor and critic share the same optimizer
        self.share_optimizer = False # if True, lr for actor and critic will be the same
        self.ppo_type = 'clip' # clip or kl
        self.eps_clip = 0.2 # clip parameter for PPO
        self.gae_lambda = 0.95 # lambda for GAE
        # for kl penalty version of PPO
        self.kl_target = 0.1 # target KL divergence
        self.kl_lambda = 0.5 # lambda for KL penalty, 0.5 is the default value in the paper
        self.kl_beta = 1.5 # beta for KL penalty, 1.5 is the default value in the paper
        self.kl_alpha = 2 # alpha for KL penalty, 2 is the default value in the paper
        self.action_type_list = "continuous" # continuous action space
        self.gamma = 0.99 # discount factor
        self.k_epochs = 4 # update policy for K epochs
        self.lr = 0.0001 # for shared optimizer
        self.actor_lr = 0.0003 # learning rate for actor, must be specified if share_optimizer is False
        self.critic_lr = 0.001 # learning rate for critic, must be specified if share_optimizer is False
        self.critic_loss_coef = 0.001 # critic loss coefficient
        self.entropy_coef = 0.01 # entropy coefficient
        self.batch_size = 256 # ppo train batch size
        self.sgd_batch_size = 32 # sgd batch size
        self.actor_hidden_dim = 256 # hidden dimension for actor
        self.critic_hidden_dim = 256 # hidden dimension for critic
        self.min_policy = 0 # min value for policy (for discrete action space)
        self.buffer_type = 'REPLAY_QUE'
        self.branch_layers = []
        self.merge_layers = []
        self.actor_branch_layers = []
        self.actor_merge_layers = []
        self.critic_branch_layers = []
        self.critic_merge_layers = []



#!/usr/bin/env python
# coding=utf-8
'''
Author: JiangJi
Email: johnjim0816@gmail.com
Date: 2023-12-22 13:55:56
LastEditor: JiangJi
LastEditTime: 2023-12-22 14:07:35
Discription: 
'''
import argparse
from joyrl import run

def main():
    parser = argparse.ArgumentParser(description="hyperparameters")
    parser.add_argument('--yaml', default=None, type=str,
                            help='the path of config file')
    args = parser.parse_args()
    run(yaml_path = args.yaml)
if __name__ == "__main__":
    main()


#!/usr/bin/env python
# coding=utf-8
'''
Author: JiangJi
Email: johnjim0816@gmail.com
Date: 2023-12-22 23:02:13
LastEditor: JiangJi
LastEditTime: 2024-06-10 23:51:23
Discription: 
'''
import torch
import torch.nn as nn
import torch.optim as optim
import torch.utils.data as Data
import numpy as np
from joyrl.algos.base.network import ActorCriticNetwork, CriticNetwork, ActorNetwork
from joyrl.algos.base.policy import BasePolicy
from joyrl.framework.config import MergedConfig

class Policy(BasePolicy):
    def __init__(self, cfg: MergedConfig) -> None:
        super(Policy, self).__init__(cfg)
        self.independ_actor = cfg.independ_actor
        self.gamma = cfg.gamma
        # self.action_type_list = cfg.action_type
        # if self.action_type_list.lower() == 'continuous': # continuous action space
        #     self.action_scale = torch.tensor((self.action_space.high - self.action_space.low)/2, device=self.device, dtype=torch.float32).unsqueeze(dim=0)
        #     self.action_bias = torch.tensor((self.action_space.high + self.action_space.low)/2, device=self.device, dtype=torch.float32).unsqueeze(dim=0)
        self.critic_loss_coef = cfg.critic_loss_coef
        self.k_epochs = cfg.k_epochs # update policy for K epochs
        self.eps_clip = cfg.eps_clip # clip parameter for PPO
        self.entropy_coef = cfg.entropy_coef # entropy coefficient
        self.sgd_batch_size = cfg.sgd_batch_size
        self.create_model()
        self.create_optimizer()
        self.create_summary()
    
    def create_summary(self):
        '''
        创建 tensorboard 数据
        '''
        self.summary = {
            'scalar': {
                'tot_loss': 0.0,
                'actor_loss': 0.0,
                'critic_loss': 0.0,
            },
        }
    def update_summary(self):
        ''' 更新 tensorboard 数据
        '''
        if hasattr(self, 'tot_loss') and len(self.tot_losses_epoch) > 0:    
            self.summary['scalar']['tot_loss'] = np.mean(self.tot_losses_epoch)
        self.summary['scalar']['actor_loss'] = np.mean(self.actor_losses_epoch)
        self.summary['scalar']['critic_loss'] = np.mean(self.critic_losses_epoch)

    def create_model(self):
        self.model = ActorCriticNetwork(self.cfg, self.state_size_list).to(self.device)

    def create_optimizer(self):
        self.optimizer = optim.Adam(self.model.parameters(), lr = self.cfg.lr)  

    def update_policy_transition(self):
        self.policy_transition = {'value': self.value.detach().cpu().numpy().item(), 'log_prob': self.log_prob}

    def sample_action(self, state, **kwargs):
        state = torch.tensor(np.array(state), device=self.device, dtype=torch.float32)
        # single state shape must be [batch_size, state_dim]
        if state.dim() == 1: 
            state = state.unsqueeze(dim=0)
        model_outputs = self.model(state)
        self.value = model_outputs['value']
        actor_outputs = model_outputs['actor_outputs']
        actions, self.log_prob = self.model.action_layers.get_actions_and_log_probs(mode = 'sample', actor_outputs = actor_outputs)
        self.update_policy_transition()
        return actions

    @torch.no_grad()
    def predict_action(self, state, **kwargs):
        state = torch.tensor(np.array(state), device=self.device, dtype=torch.float32)
        # single state shape must be [batch_size, state_dim]
        if state.dim() == 1: 
            state = state.unsqueeze(dim=0)
        model_outputs = self.model(state)
        actor_outputs = model_outputs['actor_outputs']
        actions = self.model.action_layers.get_actions(mode = 'predict', actor_outputs = actor_outputs)
        return actions
    
    def prepare_data_before_learn(self, **kwargs):
        super().prepare_data_before_learn(**kwargs)
        log_probs, returns = kwargs.get('log_probs'), kwargs.get('returns')
        self.log_probs = torch.cat(log_probs, dim=0).detach() # [batch_size,1]
        self.returns = torch.tensor(returns, dtype = torch.float32, device = self.device).unsqueeze(dim=1)
        
    def learn(self, **kwargs):
        super().learn(**kwargs)
        states, actions, log_probs, returns = kwargs.get('states'), kwargs.get('actions'), kwargs.get('log_probs'), kwargs.get('returns')

        model_outputs = self.model(states)
        values = model_outputs['value']
        advantages = returns - values.detach() # shape:[batch_size,1]
        self.actor_loss = - (log_probs * advantages).mean()
        self.critic_loss = self.cfg.critic_loss_coef * nn.MSELoss()(returns, values) # shape: [batch_size, 1]
        self.actor_losses_epoch, self.critic_losses_epoch, self.tot_losses_epoch = [], [], []
        for _ in range(self.k_epochs):
            for data in train_loader:
                # multi-head state
                old_states = []
                for i in range(len(states)):
                    old_states.append(data[i])
                idx = len(states)
                # multi-head action
                old_actions = []
                for i in range(len(actions)):
                    old_actions.append(data[idx+i])
                idx += len(actions)
                old_log_probs = data[idx]
                returns = data[idx+1]
                model_outputs = self.model(old_states)
                values = model_outputs['value']
                actor_outputs = model_outputs['actor_outputs']
                new_log_probs = self.model.action_layers.get_log_probs_action(actor_outputs, old_actions)
                # new_log_probs = self.model.action_layers.get_log_probs_action(old_actions)
                entropy_mean = self.model.action_layers.get_mean_entropy(actor_outputs)
                advantages = returns - values.detach() # shape:[batch_size,1]
                # get action probabilities
                # compute ratio (pi_theta / pi_theta__old):
                ratio = torch.exp(new_log_probs - old_log_probs) # shape: [batch_size, 1]
                # compute surrogate loss
                surr1 = ratio * advantages # shape: [batch_size, 1]
                surr2 = torch.clamp(ratio, 1 - self.eps_clip, 1 + self.eps_clip) * advantages
                # compute actor loss
                self.actor_loss = - (torch.mean(torch.min(surr1, surr2)) )
                # compute critic loss
                self.critic_loss = self.cfg.critic_loss_coef * nn.MSELoss()(returns, values) # shape: [batch_size, 1]
                self.actor_losses_epoch.append(self.actor_loss.item())
                self.critic_losses_epoch.append(self.critic_loss.item())
                self.tot_loss = self.actor_loss + self.critic_loss + self.entropy_coef * entropy_mean
                # compute total loss
                self.optimizer.zero_grad()
                self.tot_losses_epoch.append(self.tot_loss.item())
                self.tot_loss.backward()
                for param in self.model.parameters():
                    param.grad.data.clamp_(-1, 1)   
                self.optimizer.step()
        self.update_summary()
    


#!/usr/bin/env python
# coding=utf-8
'''
Author: JiangJi
Email: johnjim0816@gmail.com
Date: 2023-12-22 23:02:13
LastEditor: JiangJi
LastEditTime: 2024-06-14 09:34:14
Discription: 
'''
import time
from joyrl.framework.message import Msg, MsgType
from joyrl.framework.config import MergedConfig
from joyrl.framework.base import Moduler
from joyrl.framework.utils import exec_method

class Collector(Moduler):
    ''' Collector for collecting training data
    '''
    def __init__(self, cfg: MergedConfig, **kwargs) -> None:
        super().__init__(cfg, **kwargs)
        self.data_handler = kwargs['data_handler']
        self.training_data_que = kwargs['training_data_que']
        self.sample_data_que = kwargs['sample_data_que']
        self._get_training_data_time = time.time()
        self._consumed_exp_len = 0
        self._handle_exps_time = time.time()
        self._produced_exp_len = 0
        self._t_interval = 3
        self._t_start()

    def _t_start(self):
        exec_method(self.logger, 'info', 'get', "[Collector._t_start] Start collector!") 

    def run(self):
        sample_data_len = 0
        while True:
            exps = exec_method(self.sample_data_que, 'pop', 'get')
            if exps:
                sample_data_len += len(exps)
                self._handle_exps(exps)
            if sample_data_len >= self.cfg.batch_size // 2:
                training_data = self._get_training_data()
                if training_data:
                    exec_method(self.training_data_que, 'put', 'remote', training_data)
                    sample_data_len = 0
            
    def pub_msg(self, msg: Msg):
        ''' publish message
        '''
        msg_type, msg_data = msg.type, msg.data
        if msg_type == MsgType.COLLECTOR_PUT_EXPS:
            exps = msg_data
            self._handle_exps(exps)
        elif msg_type == MsgType.COLLECTOR_GET_TRAINING_DATA:
            return self._get_training_data()
        elif msg_type == MsgType.COLLECTOR_GET_BUFFER_LENGTH:
            return self.get_buffer_length()
        else:
            raise NotImplementedError

    def _handle_exps(self, exps):
        ''' handle exps
        '''
        exec_method(self.data_handler, 'add_exps', 'remote', exps)
        self._produced_exp_len += len(exps)
        if time.time() - self._handle_exps_time >= self._t_interval:
            exec_method(self.logger, 'info', 'remote', f"[Collector._handle_exps] SAMPLE PRODUCTION SPEED per second: {self._produced_exp_len/self._t_interval:.2f}")
            self._handle_exps_time = time.time()
            self._produced_exp_len = 0

    def _get_training_data(self):
        ''' get training data
        '''
        get_training_data_time = time.time()
        training_data = None
        while True:
            training_data = exec_method(self.data_handler, 'get_training_data', 'get')
            if training_data:
                self._consumed_exp_len += training_data['dones'].shape[0]
                break
            if self.cfg.is_learner_async:
                break
            if time.time() - get_training_data_time >= 0.02:
                # exec_method(self.logger, 'warning', 'remote', "[Collector._get_training_data] get training data timeout!")
                get_training_data_time = time.time()
                break
        if time.time() - self._get_training_data_time >= self._t_interval:
            exec_method(self.logger, 'info', 'remote', f"[Collector.pub_msg] SAMPLE CONSUMPTION SPEED per second: {self._consumed_exp_len/self._t_interval:.2f}")
            self._get_training_data_time = time.time()
            self._consumed_exp_len = 0
        return training_data
    


#!/usr/bin/env python
# coding=utf-8
'''
Author: JiangJi
Email: johnjim0816@gmail.com
Date: 2023-12-02 15:02:30
LastEditor: JiangJi
LastEditTime: 2024-06-14 09:33:01
Discription: 
'''
import copy
import time
import ray
from ray.util.queue import Queue as RayQueue
from joyrl.framework.message import Msg, MsgType
from joyrl.framework.config import MergedConfig
from joyrl.framework.base import Moduler
from joyrl.framework.collector import Collector
from joyrl.framework.tracker import Tracker
from joyrl.framework.interactor import Interactor
from joyrl.framework.learner import Learner
from joyrl.framework.tester import OnlineTester
from joyrl.framework.policy_mgr import PolicyMgr
from joyrl.framework.recorder import Recorder
from joyrl.framework.utils import exec_method, create_module
from joyrl.framework.utils import Logger
from joyrl.framework.utils import DataActor, QueueActor

class Trainer(Moduler):
    ''' Collector for collecting training data
    '''
    def __init__(self, cfg: MergedConfig,**kwargs) -> None:
        super().__init__(cfg, **kwargs)
        self.env = kwargs['env']
        self.policy = kwargs['policy']
        self.data_handler = kwargs['data_handler']
        self._print_cfgs() # print parameters
        self._create_shared_data() # create data queues
        self._create_modules() # create modules
        
    def _create_shared_data(self):
        self.latest_model_params_dict = DataActor.options(**{'num_cpus': 0}).remote({'step': 0, 'model_params': self.policy.get_model_params()})
        self.sample_data_que = QueueActor.remote(maxsize=256)
        self.training_data_que = QueueActor.remote(maxsize=1)
        # self.sample_data_que = RayQueue(maxsize=256)
        # self.training_data_que = RayQueue(maxsize=1)
        
    def _create_modules(self):
        ''' create modules
        '''
        self.policy_mgr = ray.remote(PolicyMgr).options(**{'num_cpus': 0}).remote(
            self.cfg,
            name = 'PolicyMgr',
            policy = copy.deepcopy(self.policy),
            latest_model_params_dict = self.latest_model_params_dict,
        )
        if self.cfg.online_eval:
            recorder = ray.remote(Recorder).options(**{'num_cpus': 0}).remote(self.cfg,
                                                                            name = 'RecorderOnlineTester',
                                                                            type = 'online_tester')
            self.online_tester = OnlineTester(
                self.cfg, 
                name = 'OnlineTester',
                env = copy.deepcopy(self.env), 
                policy = copy.deepcopy(self.policy),
                recorder = recorder,
                policy_mgr = self.policy_mgr,
            )
        self.tracker = ray.remote(Tracker).remote(self.cfg)
        self.collector = ray.remote(Collector).options(**{'num_cpus': 1}).remote(
            self.cfg,
            name = 'Collector',
            data_handler = self.data_handler,
            sample_data_que = self.sample_data_que,
            training_data_que = self.training_data_que,
        )
        self.interactors = []
        recorder = ray.remote(Recorder).options(**{'num_cpus': 0}).remote(self.cfg,
                                                                            name = 'Recorder_Interactor',
                                                                            type = 'interactor')
        for i in range(self.cfg.n_interactors):
            interactor = ray.remote(Interactor).options(**{'num_cpus': 1}).remote(
                self.cfg,
                id = i,
                name = f"Interactor_{i}",
                env = copy.deepcopy(self.env),
                policy = copy.deepcopy(self.policy),
                data_handler = copy.deepcopy(self.data_handler),  # only use the static method handle_exps_after_interact
                tracker = self.tracker,
                collector = self.collector,
                recorder = recorder,
                policy_mgr = self.policy_mgr,
                latest_model_params_dict = self.latest_model_params_dict,
                sample_data_que = self.sample_data_que,
                )
            self.interactors.append(interactor)
        self.learners = []
        recorder = ray.remote(Recorder).options(**{'num_cpus': 0}).remote(self.cfg,
                                                                            name = 'Recorder_Learner',
                                                                            type = 'learner')
        for i in range(self.cfg.n_learners):
            learner = ray.remote(Learner).remote(
                self.cfg,
                id = i,
                name = f"Learner_{i}",
                policy = copy.deepcopy(self.policy),
                policy_mgr = self.policy_mgr,
                collector = self.collector,
                data_handler = self.data_handler,
                tracker = self.tracker,
                recorder = recorder,
                training_data_que = self.training_data_que,
                )
            self.learners.append(learner)

    def _print_cfgs(self):
        ''' print parameters
        '''
        def print_cfg(cfg, name = ''):
            cfg_dict = vars(cfg)
            exec_method(self.logger, 'info', 'get', f"{name}:")
            exec_method(self.logger, 'info', 'get', ''.join(['='] * 80))
            tplt = "{:^20}\t{:^20}\t{:^20}"
            exec_method(self.logger, 'info', 'get', tplt.format("Name", "Value", "Type"))
            for k, v in cfg_dict.items():
                if v.__class__.__name__ == 'list': # convert list to str
                    v = str(v)
                if v is None: # avoid NoneType
                    v = 'None'
                if "support" in k: # avoid ndarray
                    v = str(v[0])
                exec_method(self.logger, 'info', 'get', tplt.format(k, v, str(type(v))))
            exec_method(self.logger, 'info', 'get', ''.join(['='] * 80))
        print_cfg(self.cfg.general_cfg, name = 'General Configs')
        print_cfg(self.cfg.algo_cfg, name = 'Algo Configs')
        print_cfg(self.cfg.env_cfg, name = 'Env Configs')

    def run(self):
        ''' run the trainer
        '''
        exec_method(self.logger, 'info', 'get', f"[Trainer.run] Start {self.cfg.mode}ing!")
        s_t = time.time()
        exec_method(self.tracker, 'pub_msg', 'get', Msg(type = MsgType.TRACKER_INCREASE_UPDATE_STEP, data = self.cfg.start_model_step))
        if not self.cfg.is_learner_async:
            while True:
                ray.get([interactor.run.remote() for interactor in self.interactors])
                ray.get([learner.run.remote() for learner in self.learners])
                if exec_method(self.tracker, 'pub_msg', 'get', Msg(type = MsgType.TRACKER_CHECK_TASK_END)):
                    e_t = time.time()
                    exec_method(self.logger, 'info', 'get', f"[Trainer.run] Finish {self.cfg.mode}ing! Time cost: {e_t - s_t:.3f} s")
                    time.sleep(5)
                    ray.shutdown()
                    break
        else:
            self.collector.run.remote()
            [interactor.run.remote() for interactor in self.interactors]
            [learner.run.remote() for learner in self.learners]
            while True:
                if exec_method(self.tracker, 'pub_msg', 'get', Msg(type = MsgType.TRACKER_CHECK_TASK_END)):
                    e_t = time.time()
                    exec_method(self.logger, 'info', 'get', f"[Trainer.run] Finish {self.cfg.mode}ing! Time cost: {e_t - s_t:.3f} s")
                    time.sleep(5)
                    ray.shutdown()
                    break
                time.sleep(1)


#!/usr/bin/env python
# coding=utf-8
'''
Author: JiangJi
Email: johnjim0816@gmail.com
Date: 2023-04-28 16:18:44
LastEditor: JiangJi
LastEditTime: 2024-06-14 09:34:29
Discription: 
'''
from ray.util.queue import Queue as RayQueue
from pathlib import Path
import pickle
import time
import threading

import pandas as pd
from torch.utils.tensorboard import SummaryWriter  
from joyrl.framework.message import Msg, MsgType
from joyrl.framework.config import MergedConfig
from joyrl.framework.base import Moduler
from joyrl.framework.utils import exec_method

class Recorder(Moduler):
    ''' Recorder for recording training information
    '''
    def __init__(self, cfg: MergedConfig, **kwargs) -> None:
        super().__init__(cfg, **kwargs)
        self.type = kwargs.get('type', 'recorder')
        self.writter = SummaryWriter(log_dir=f"{self.cfg.tb_dir}/{self.type}")
        self._summary_que = RayQueue(maxsize = 256)
        self._t_start() # TODO, slow down the training speed when using threading 

    def _t_start(self):
        exec_method(self.logger, 'info', 'remote', f"[Recorder._t_start] Start {self.type} recorder!")
        self._t_save_summary = threading.Thread(target=self._save_summary)
        self._t_save_summary.setDaemon(True)
        self._t_save_summary.start()

    def pub_msg(self, msg: Msg):
        ''' publish message
        '''
        msg_type, msg_data = msg.type, msg.data
        if msg_type == MsgType.RECORDER_PUT_SUMMARY:
            summary_data_list = msg_data
            self._add_summary(summary_data_list)
        else:
            raise NotImplementedError
        
    def _init_writter(self):
        self.writters = {}
        self.writter_types = ['interact','policy']
        for writter_type in self.writter_types:
            self.writters[writter_type] = SummaryWriter(log_dir=f"{self.cfg.tb_dir}/{writter_type}")
    
    def _add_summary(self, summary_data_list):
        while True:
            try:
                self._summary_que.put(summary_data_list, block = False)
                break
            except:
                self.logger.warning(f"[Recorder._add_summary] {self.type} summary_que is full!")
                # time.sleep(0.001)
                pass

    def _write_tb_scalar(self, step: int, summary: dict):
        for key, value in summary.items():
            self.writter.add_scalar(tag = f"{self.cfg.mode.lower()}_{key}", scalar_value = value, global_step = step)

    def _write_dataframe(self, step: int, summary: dict):
        df_file = f"{self.cfg.res_dir}/{self.type}.csv"
        if Path(df_file).exists():
            df = pd.read_csv(df_file)
        else:
            df = pd.DataFrame()
        saved_dict = {f"{self.type}_step": step}
        saved_dict.update(summary)
        df = pd.concat([df, pd.DataFrame(saved_dict, index=[0])], ignore_index = True)
        df.to_csv(df_file, index = False)

    def _save_summary(self):
        while True:
            while not self._summary_que.empty():
                summary_data_list = self._summary_que.get()
                for summary_data in summary_data_list:
                    step, summary = summary_data
                    self._write_tb_scalar(step, summary)
                    self._write_dataframe(step, summary)
            time.sleep(0.001)

class BaseTrajCollector:
    ''' Base class for trajectory collector
    '''
    def __init__(self, fpath) -> None:
        pass
class SimpleTrajCollector(BaseTrajCollector):
    ''' Simple trajectory collector for store trajectories
    '''
    def __init__(self, fpath) -> None:
        super().__init__(fpath)
        self.fpath = fpath
        self.traj_num = 0
        self.init_traj()
        self.init_traj_cache()
    def init_traj(self):
        ''' init trajectories
        '''
        self.trajs = {'state':[],'action':[],'reward':[],'next_state':[],'terminated':[],'info':[]}
    def init_traj_cache(self):
        ''' init trajectory cache for one episode
        '''
        self.ep_states, self.ep_actions, self.ep_rewards, self.ep_next_states, self.ep_terminated, self.ep_infos = [], [], [], [], [], []
    def add_traj_cache(self,state,action,reward,next_state,terminated,info):
        ''' store one episode trajectory
        '''
        self.ep_states.append(state)
        self.ep_actions.append(action)
        self.ep_rewards.append(reward)
        self.ep_next_states.append(next_state)
        self.ep_terminated.append(terminated)
        self.ep_infos.append(info)
    def store_traj(self, task_end_flag = False):
        ''' store trajectory cache into trajectories
        '''
        self.trajs['state'].append(self.ep_states)
        self.trajs['action'].append(self.ep_actions)
        self.trajs['reward'].append(self.ep_rewards)
        self.trajs['next_state'].append(self.ep_next_states)
        self.trajs['terminated'].append(self.ep_terminated)
        self.trajs['info'].append(self.ep_infos)
        if len(self.trajs['state']) >= 1000 or task_end_flag: # save traj when traj number is greater than 1000
            with open(f"{self.fpath}/trajs_{self.traj_num}.pkl", 'wb') as f:
                pickle.dump(self.trajs, f)
                self.traj_num += 1
            self.init_traj_cache()


#!/usr/bin/env python
# coding=utf-8
'''
Author: JiangJi
Email: johnjim0816@gmail.com
Date: 2023-05-17 01:08:36
LastEditor: JiangJi
LastEditTime: 2024-06-10 23:51:36
Discription: 
'''
import numpy as np
import torch
from joyrl.algos.base.data_handler import BaseDataHandler
class DataHandler(BaseDataHandler):
    def __init__(self, cfg):
        super().__init__(cfg)
        self.gae_lambda = getattr(self.cfg, 'gae_lambda', 0.95)
        self.gamma = getattr(self.cfg, 'gamma', 0.95)
        self.batch_exps = []
    
    def handle_exps_after_interact(self, exps):
        exp_len = self._get_exp_len(exps)
        next_value = exps[-1].value
        return_mc = 0
        return_td = next_value   
        adv_gae = 0
        returns_mc = []
        returns_td = []
        returns_gae = []
        for t in reversed(range(exp_len)):
            delta = exps[t].reward + self.gamma * next_value * (1 - exps[t].done) - exps[t].value
            adv_gae = delta + self.gamma * self.gae_lambda * (1 - exps[t].done) * adv_gae
            return_mc = exps[t].reward + self.gamma * return_mc * (1 - exps[t].done)
            return_td = exps[t].reward + self.gamma * return_td * (1 - exps[t].done)
            returns_mc.insert(0, return_mc)
            returns_td.insert(0, return_td)
            returns_gae.insert(0, adv_gae + exps[t].value)
            exps[t].return_mc = return_mc
            exps[t].return_td = return_td
            exps[t].adv_gae = adv_gae
            exps[t].return_gae = adv_gae + exps[t].value
            next_value = exps[t].value
        return_mc_normed = (returns_mc - np.mean(returns_mc)) / (np.std(returns_mc) + 1e-8)
        return_td_normed = (returns_td - np.mean(returns_td)) / (np.std(returns_td) + 1e-8)
        return_gae_normed = (returns_gae - np.mean(returns_gae)) / (np.std(returns_gae) + 1e-8)
        for t in range(exp_len):
            exps[t].return_mc_normed = return_mc_normed[t]
            exps[t].normed_return_td = return_td_normed[t]
            exps[t].normed_return_gae = return_gae_normed[t]
        exps = exps[:exp_len]
        return exps
    
    def add_exps(self, exps):
        exps = self.handle_exps_after_interact(exps)
        self.batch_exps.extend(exps)
        if len(self.batch_exps) >= self.cfg.batch_size:
            self.buffer.push(self.batch_exps)
            self.batch_exps = []

    def _handle_exps_before_train(self, exps: list):
        ''' convert exps to training data
        '''
        super()._handle_exps_before_train(exps)
        log_probs = [exp.log_prob for exp in exps] 
        returns = np.array([exp.normed_return_td for exp in exps]) # [batch_size]
        self.data_after_train.update({'log_probs': log_probs, 'returns': returns})

        


#!/usr/bin/env python
# coding=utf-8
'''
Author: JiangJi
Email: johnjim0816@gmail.com
Date: 2023-12-22 23:02:13
LastEditor: JiangJi
LastEditTime: 2023-12-24 22:52:08
Discription: 
'''
import gymnasium as gym
from gymnasium.envs.registration import register

def register_env(env_name):
    if env_name == 'Racetrack-v0':
        register(
            id='Racetrack-v0',
            entry_point='envs.racetrack:RacetrackEnv',
            max_episode_steps=1000,
            kwargs={}
        )
    elif env_name == 'FrozenLakeNoSlippery-v1':
        register(
            id='FrozenLakeNoSlippery-v1',
            entry_point='gym.envs.toy_text.frozen_lake:FrozenLakeEnv',
            kwargs={'map_name':"4x4",'is_slippery':False},
        )
    elif env_name == 'CustomCliffWalking-v0':
        register(
            id='CustomCliffWalking-v0',
            entry_point='joyrl.envs.gym.toy_text.cliff_walking:CustomCliffWalkingEnv',
            max_episode_steps=1000,
            kwargs={}
        )
    else:
        pass

# if __name__ == "__main__":
#     import random
#     import gym
#     env = gym.make('FrozenLakeNoSlippery-v1')
#     num_steps = 1000000
#     state = env.reset()
#     n_actions = env.action_space.n
#     print(state)
#     for _ in range(num_steps) :
#         next_state, reward, done,_ = env.step(random.choice(range(n_actions)))
#         print(next_state)
#         if (done) :
#             _ = env.reset()
    



import turtle
import gymnasium as gym_wrapper
import torch
import numpy as np
from torchvision import transforms
from gymnasium.spaces import Box
from gymnasium.wrappers import FrameStack
from typing import List


class BaseSkipFrame(gym_wrapper.Wrapper):
    def __init__(
            self, 
            env, 
            skip: int, 
            cut_slices: List[List[int]]=None,
            start_skip: int=None,
            int_action_flag: bool=False
        ):
        """_summary_
        Args:
            env (_type_): _description_
            skip (int): skip frames
            cut_slices (List[List[int]], optional): pic observation cut. Defaults to None.
            start_skip (int, optional): skip several frames to start. Defaults to None.
            int_action_flag (bool): if the action only a int,  set true. Defaults to False.
        """
        super().__init__(env)
        self._skip = skip
        self.pic_cut_slices = cut_slices
        self.start_skip = start_skip
        self.int_action_flag = int_action_flag

    def _cut_slice(self, obs):
        slice_list = []
        for idx, dim_i_slice in enumerate(self.pic_cut_slices):
            slice_list.append(eval('np.s_[{st}:{ed}]'.format(st=dim_i_slice[0], ed=dim_i_slice[1])))

        obs = obs[tuple(i for i in slice_list)]
        return obs

    def step(self, action):
        tt_reward_list = []
        done = False
        total_reward = 0
        if self.int_action_flag:
            action = action[0]
        for i in range(self._skip):
            obs, reward, terminated, truncated, info = self.env.step(action)
            done_f = terminated or truncated
            total_reward += reward
            tt_reward_list.append(reward)
            if done:
                break

        obs = self._cut_slice(obs)  if self.pic_cut_slices is not None else obs
        return obs, total_reward, done_f, truncated, info

    def _start_skip(self):
        a = np.array([0.0, 0.0, 0.0]) if hasattr(self.env.action_space, 'low') else np.array(0) 
        for i in range(self.start_skip):
            obs, reward, terminated, truncated, info = self.env.step(a)
        return obs, info

    def reset(self, seed=0, options=None):
        s, info = self.env.reset(seed=seed, options=options)
        if self.start_skip is not None:
            obs, info = self._start_skip()
        obs = self._cut_slice(obs)  if self.pic_cut_slices is not None else obs
        return obs, info


class GymDiscreteActionWrapper(gym_wrapper.Wrapper):
    def __init__(self, env):
        gym_wrapper.Wrapper.__init__(self, env)
        # self.action_space = Box(low=-1.0, high=1.0, shape=(1, ), dtype=np.float32)

    def step(self, action):
        action = action[0]
        return self.env.step(action)

    def reset(self, seed=0, options=None):
        return self.env.reset(seed=seed, options=options)
    
class CliffWalkingWapper(gym_wrapper.Wrapper):
    def __init__(self, env):
        gym_wrapper.Wrapper.__init__(self, env)
        self.t = None
        self.unit = 50
        self.max_x = 12
        self.max_y = 4

    def draw_x_line(self, y, x0, x1, color='gray'):
        assert x1 > x0
        self.t.color(color)
        self.t.setheading(0)
        self.t.up()
        self.t.goto(x0, y)
        self.t.down()
        self.t.forward(x1 - x0)

    def draw_y_line(self, x, y0, y1, color='gray'):
        assert y1 > y0
        self.t.color(color)
        self.t.setheading(90)
        self.t.up()
        self.t.goto(x, y0)
        self.t.down()
        self.t.forward(y1 - y0)

    def draw_box(self, x, y, fillcolor='', line_color='gray'):
        self.t.up()
        self.t.goto(x * self.unit, y * self.unit)
        self.t.color(line_color)
        self.t.fillcolor(fillcolor)
        self.t.setheading(90)
        self.t.down()
        self.t.begin_fill()
        for i in range(4):
            self.t.forward(self.unit)
            self.t.right(90)
        self.t.end_fill()

    def move_player(self, x, y):
        self.t.up()
        self.t.setheading(90)
        self.t.fillcolor('red')
        self.t.goto((x + 0.5) * self.unit, (y + 0.5) * self.unit)

    def render(self):
        if self.t == None:
            self.t = turtle.Turtle()
            self.wn = turtle.Screen()
            self.wn.setup(self.unit * self.max_x + 100,
                          self.unit * self.max_y + 100)
            self.wn.setworldcoordinates(0, 0, self.unit * self.max_x,
                                        self.unit * self.max_y)
            self.t.shape('circle')
            self.t.width(2)
            self.t.speed(0)
            self.t.color('gray')
            for _ in range(2):
                self.t.forward(self.max_x * self.unit)
                self.t.left(90)
                self.t.forward(self.max_y * self.unit)
                self.t.left(90)
            for i in range(1, self.max_y):
                self.draw_x_line(
                    y=i * self.unit, x0=0, x1=self.max_x * self.unit)
            for i in range(1, self.max_x):
                self.draw_y_line(
                    x=i * self.unit, y0=0, y1=self.max_y * self.unit)

            for i in range(1, self.max_x - 1):
                self.draw_box(i, 0, 'black')
            self.draw_box(self.max_x - 1, 0, 'yellow')
            self.t.shape('turtle')

        x_pos = self.s % self.max_x
        y_pos = self.max_y - 1 - int(self.s / self.max_x)
        self.move_player(x_pos, y_pos)



class CarV2SkipFrame(gym_wrapper.Wrapper):
    def __init__(self, env, skip: int = 5, continue_flag: bool=False):
        """skip frame
        Args:
            env (_type_): _description_
            skip (int): skip frames
            continue_flag (bool): continuous flag in carRacing env 
        """
        super().__init__(env)
        self._skip = skip
        self.continue_flag = continue_flag
    
    def step(self, action):
        tt_reward_list = []
        done = False
        total_reward = 0
        for i in range(self._skip):
            obs, reward, done, info, _ = self.env.step(action)
            out_done = self.judge_out_of_route(obs)
            # done_f = done
            done_f = done or out_done
            reward = -10 if out_done else reward
            total_reward += reward
            tt_reward_list.append(reward)
            if done_f:
                break
        return obs[:84, 6:90, :], total_reward, done_f, info, _
    
    def judge_out_of_route(self, obs):
        s = obs[:84, 6:90, :]
        out_sum = (s[75, 35:48, 1][:2] > 200).sum() + (s[75, 35:48, 1][-2:] > 200).sum()
        return out_sum == 4

    def reset(self, seed=0, options=None):
        s, info = self.env.reset(seed=seed, options=options)
        # continue
        a = 0
        if self.continue_flag:
            a = np.array([0.0, 0.0, 0.0]) 
        for i in range(45):
            obs, reward, done, info, _ = self.env.step(a)

        return obs[:84, 6:90, :], info


class GrayScaleObservation(gym_wrapper.ObservationWrapper):
    def __init__(self, env):
        """RGP -> Gray
        (high, width, channel) -> (1, high, width) 
        """
        super().__init__(env)
        self.observation_space = Box(
            low=0, high=255, shape=self.observation_space.shape[:2], dtype=np.uint8
        )
    
    def observation(self, observation):
        tf = transforms.Grayscale()
        # channel first
        return tf(torch.tensor(np.transpose(observation, (2, 0, 1)).copy(), dtype=torch.float))


class ResizeObservation(gym_wrapper.ObservationWrapper):
    def __init__(self, env, shape: int = 84):
        """reshape observe
        Args:
            env (_type_): _description_
            shape (int): reshape size
        """
        super().__init__(env)
        self.shape = (shape, shape)
        obs_shape = self.shape + self.observation_space.shape[2:]
        self.observation_space = Box(low=0, high=255, shape=obs_shape, dtype=np.uint8)

    def observation(self, observation):
        #  Normalize -> input[channel] - mean[channel]) / std[channel]
        tf = transforms.Compose([transforms.Resize(self.shape), transforms.Normalize(0, 255)])
        return tf(observation).squeeze(0)

# 跳帧
class SkipFrame(gym_wrapper.Wrapper):
    def __init__(self, env, skip: int = 5, skip_start: int = 0):
        """skip frame
        Args:
            env (_type_): _description_
            skip (int): skip frames
            skip_start (int): skip start frames
        """
        super().__init__(env)
        self._skip = skip
        self._skip_start = skip_start
    
    def step(self, action):
        tt_reward_list = []
        done = False
        total_reward = 0
        for i in range(self._skip):
            if type(action) != int:
                action = action[0]
            obs, reward, done, truncated, info = self.env.step(action)
            total_reward += reward
            tt_reward_list.append(reward)
            if done:
                break
        return obs, total_reward, done, truncated, info

    def reset(self, seed=0, options=None):
        obs, info = self.env.reset(seed=seed, options=options)
        action = 0
        for i in range(self._skip_start):
            obs, reward, done, truncated, info = self.env.step(action)

        return obs, info

# 裁剪
class CropFrame(gym_wrapper.ObservationWrapper):
    def __init__(self, env, x1=0, x2=0, y1=0, y2=0):
        """crop frame
        Args:
            env (_type_): _description_
            x1 (int): crop left
            x2 (int): crop right
            y1 (int): crop top
            y2 (int): crop bottom
        """
        super().__init__(env)
        self.x1 = x1
        self.x2 = x2
        self.y1 = y1
        self.y2 = y2
    
    def observation(self, obs):
        # save_img(obs)
        if self.x1 or self.x2:
            obs = obs[:, self.x1:self.x2, :]
        if self.y1 or self.y2:
            obs = obs[self.y1:self.y2, :, :]
        return obs

class InfoRewardFrame(gym_wrapper.Wrapper):
    def __init__(self, env, goods=[], bads=[]):
        """skip frame
        Args:
            env (_type_): _description_
            skip (int): skip frames
            skip_start (int): skip start frames
        """
        super().__init__(env)
        self._goods = goods
        self._bads = bads
        self.last_info = None
    
    def step(self, action):
        obs, reward, done, truncated, info = self.env.step(action)
        if not self.last_info:
            self.last_info = info
        for good in self._goods:
            rate = self._goods[good]
            reward += (info[good] - self.last_info[good]) * rate
        for bad in self._bads:
            rate = self._bads[bad]
            reward += (info[bad] - self.last_info[bad]) * -rate
        self.last_info = info
        # print(info, reward)
        return obs, reward, done, truncated, info

    def reset(self, seed=0, options=None):
        self.last_info = None
        obs, info = self.env.reset(seed=seed, options=options)
        return obs, info


#!/usr/bin/env python
# coding=utf-8
'''
Author: JiangJi
Email: johnjim0816@gmail.com
Date: 2023-12-22 23:02:13
LastEditor: JiangJi
LastEditTime: 2024-06-14 09:34:00
Discription: 
'''
import time
import copy
import threading
from ray.util.queue import Queue as RayQueue
from joyrl.framework.message import Msg, MsgType
from joyrl.framework.config import MergedConfig
from joyrl.framework.base import Moduler
from joyrl.framework.utils import exec_method, load_model_meta, save_model_meta

class PolicyMgr(Moduler):
    ''' model manager
    '''
    def __init__(self, cfg: MergedConfig, *args, **kwargs) -> None:
        super().__init__(cfg, *args, **kwargs)
        self.policy = copy.deepcopy(kwargs['policy'])
        self._latest_model_params_dict = kwargs.get('latest_model_params_dict', None)
        self._saved_model_que = RayQueue(maxsize = 64)
        self._model_meta_que = RayQueue(maxsize = 64)
        self._t_start()
        
    def _t_start(self):
        exec_method(self.logger, 'info', 'get', "[PolicyMgr._t_start] Start policy manager!")
        self._t_save_policy = threading.Thread(target=self._save_policy)
        self._t_save_policy.setDaemon(True)
        self._t_save_policy.start()
        self._t_save_model_meta = threading.Thread(target=self._save_model_meta)
        self._t_save_model_meta.setDaemon(True)
        self._t_save_model_meta.start()
    
    def pub_msg(self, msg: Msg):
        ''' publish message
        '''
        msg_type, msg_data = msg.type, msg.data
        if msg_type == MsgType.POLICY_MGR_PUT_MODEL_PARAMS:
            self._put_model_params(msg_data)
        elif msg_type == MsgType.POLICY_MGR_PUT_MODEL_META: # not used in this version
            self._put_model_meta(msg_data)
        else:
            raise NotImplementedError
        
    def _put_model_params(self, msg_data):
        ''' put model params
        '''
        update_step, model_params = msg_data
        exec_method(self._latest_model_params_dict, 'set_value', 'get', {'step': update_step, 'model_params': model_params})
        if update_step % self.cfg.model_save_fre == 0:
            while True:
                try: # if queue is full, wait for 0.01s
                    self._saved_model_que.put((update_step, model_params), block=False)
                    break
                except:
                    exec_method(self.logger, 'warning', 'get', "[PolicyMgr._put_model_params] saved_model_que is full!")
                    # time.sleep(0.001)

    def _put_model_meta(self, msg_data):
        ''' put model meta
        ''' 
        model_meta = msg_data
        try: # if queue is full, wait for 0.01s
            self._model_meta_que.put(model_meta, block=False)
        except:
            exec_method(self.logger, 'warning', 'get', "[PolicyMgr._put_model_meta] model_meta_que is full!")        

    def _save_policy(self):
        ''' async run
        '''
        while True:
            while not self._saved_model_que.empty():
                update_step, model_params = self._saved_model_que.get()
                self.policy.put_model_params(model_params)
                self.policy.save_model(f"{self.cfg.model_dir}/{update_step}")
            time.sleep(0.01)

    def _save_model_meta(self):
        ''' save model meta
        '''
        while True:
            while not self._model_meta_que.empty():
                model_meta_type, model_meta = self._model_meta_que.get()
                old_model_meta = load_model_meta(f"{self.cfg.model_dir}")
                if old_model_meta.get(model_meta_type, None) is None:
                    old_model_meta[model_meta_type] = {}
                old_model_meta[model_meta_type].update(model_meta)
                save_model_meta(f"{self.cfg.model_dir}", old_model_meta)
            time.sleep(0.002)



import torch
import math
class SharedAdam(torch.optim.Adam):
    """Implements Adam algorithm with shared states.
    """

    def __init__(self,
                 params,
                 lr=1e-3,
                 betas=(0.9, 0.999),
                 eps=1e-8,
                 weight_decay=0):
        super(SharedAdam, self).__init__(params, lr, betas, eps, weight_decay)

        for group in self.param_groups:
            for p in group['params']:
                state = self.state[p]
                state['step'] = torch.zeros(1)
                state['exp_avg'] = p.data.new().resize_as_(p.data).zero_()
                state['exp_avg_sq'] = p.data.new().resize_as_(p.data).zero_()

    def share_memory(self):
        for group in self.param_groups:
            for p in group['params']:
                state = self.state[p]
                state['step'].share_memory_()
                state['exp_avg'].share_memory_()
                state['exp_avg_sq'].share_memory_()

    def step(self, closure=None):
        """Performs a single optimization step.
        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data
                state = self.state[p]

                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                beta1, beta2 = group['betas']

                state['step'] += 1

                if group['weight_decay'] != 0:
                    grad = grad.add(group['weight_decay'], p.data)

                # Decay the first and second moment running average coefficient
                exp_avg.mul_(beta1).add_(grad,alpha = 1 - beta1)
                exp_avg_sq.mul_(beta2).addcmul_(grad, grad,value = 1 - beta2)

                denom = exp_avg_sq.sqrt().add_(group['eps'])

                bias_correction1 = 1 - beta1 ** state['step'].item()
                bias_correction2 = 1 - beta2 ** state['step'].item()
                step_size = group['lr'] * math.sqrt(
                    bias_correction2) / bias_correction1

                p.data.addcdiv_(exp_avg, denom,value = -step_size)
        return loss


import copy, dill
import torch
import torch.nn as nn
import torch.optim as optim
import numpy as np
from collections import defaultdict
from gymnasium.spaces import Box, Discrete
from joyrl.algos.base.action_layer import ActionLayerType
from joyrl.framework.config import MergedConfig
from joyrl.algos.base.network import BaseNework
class BasePolicy(object):
    ''' base policy for DRL
    '''
    def __init__(self, cfg : MergedConfig) -> None:
        self.cfg = cfg
        self.device = torch.device(cfg.device) 
        self.obs_space = cfg.obs_space
        self.action_space = cfg.action_space
        self.policy_transition = {}
        self.data_after_train = {}
        self.model_meta = {}
        self.get_state_size()
        self.get_action_size()
        self.create_model()
        self.create_optimizer()
        self.create_summary()
    
    def get_state_size(self):
        ''' get state size
        '''
        # state_size must be [[None, state_dim_1], [None, state_dim_2], ...]
        if isinstance(self.obs_space, Box):
            if len(self.obs_space.shape) == 3:
                self.state_size_list = [[None, self.obs_space.shape[0], self.obs_space.shape[1], self.obs_space.shape[2]]]
            else:
                self.state_size_list = [[None, self.obs_space.shape[0]]]
        elif isinstance(self.obs_space, Discrete):
            self.state_size_list = [[None, self.obs_space.n]]
        else:
            raise ValueError('obs_space type error')
        setattr(self.cfg, 'state_size_list', self.state_size_list)
        return self.state_size_list
    
    def get_action_size(self):
        ''' get action size
        '''
        # action_size must be [action_dim_1, action_dim_2, ...]
        if isinstance(self.action_space, Box):
            self.action_size_list = [self.action_space.shape[0]]
            self.action_type_list = ["CONTINUOUS"]
            self.action_high_list = [self.action_space.high[0]]
            self.action_low_list = [self.action_space.low[0]]  
        elif isinstance(self.action_space, Discrete):
            self.action_size_list = [self.action_space.n]
            self.action_type_list = ["DISCRETE"]
            self.action_high_list = [self.action_space.n]
            self.action_low_list = [0]
        else:
            raise ValueError('action_space type error')
        setattr(self.cfg, 'action_size_list', self.action_size_list)
        setattr(self.cfg, 'action_type_list', self.action_type_list)
        setattr(self.cfg, 'action_high_list', self.action_high_list)
        setattr(self.cfg, 'action_low_list', self.action_low_list)
    
    def create_model(self):
        ''' create model
        '''
        self.model = BaseNework(self.cfg, [None, 1])

    def create_optimizer(self):
        self.optimizer = optim.Adam(self.model.parameters(), lr=self.cfg.lr) 

    def get_model_params(self):
        ''' get model params
        '''
        return self.model.state_dict()
    
    def put_model_params(self, model_params):
        ''' put model params
        '''
        self.model.load_state_dict(model_params)

    def get_action(self, state, **kwargs):
        ''' get action
        '''
        mode = kwargs.get('mode', None)
        if mode is None: 
            mode = 'sample' if self.cfg.mode == 'train' else 'predict'
        if mode == 'sample':
            return self.sample_action(state, **kwargs)
        elif mode == 'predict':
            return self.predict_action(state, **kwargs)
        else:
            raise NameError('[get_action] mode must be sample or predict')
        
    def sample_action(self, state, **kwargs):
        ''' sample action
        '''
        raise NotImplementedError
    def predict_action(self, state, **kwargs):
        ''' predict action
        '''
        raise NotImplementedError
    
    def update_policy_transition(self):
        ''' update policy transition
        '''
        self.policy_transition = {}
        
    def get_policy_transition(self):
        return self.policy_transition
    
    def create_summary(self):
        ''' create policy summary
        '''
        self.summary = {}
        self.summary['scalar'] = {}
        for i in range(len(self.action_size_list)):
            self.summary['scalar'][f'loss_{i}'] = 0.0

    def update_summary(self):
        ''' update policy summary
        '''
        for i in range(len(self.action_size_list)):
            self.summary['scalar'][f'loss_{i}'] = self.summary_loss[i]

    def get_summary(self):
        return self.summary['scalar']
    
    def learn(self, **kwargs):
        ''' learn policy
        '''
        self.prepare_data_before_learn(**kwargs)
    
    def prepare_data_before_learn(self, **kwargs):
        ''' prepare data before training
        '''
        states, actions, next_states, rewards, dones = kwargs.get('states'), kwargs.get('actions'), kwargs.get('next_states'), kwargs.get('rewards'), kwargs.get('dones')
        # multi-head state
        self.states = [ torch.tensor(states, dtype = torch.float32, device = self.cfg.device) ]
        # multi-head action
        self.actions = [ torch.tensor(actions, dtype = torch.float32, device = self.cfg.device) ]
        self.rewards = torch.tensor(rewards, dtype = torch.float32, device = self.cfg.device).unsqueeze(dim=1)
        self.next_states = torch.tensor(next_states, dtype = torch.float32, device = self.cfg.device)
        self.dones = torch.tensor(dones, dtype = torch.float32, device = self.cfg.device).unsqueeze(dim=1)

    def update_data_after_learn(self):
        ''' update data after training
        '''
        self.data_after_train = {}
        
    def get_data_after_learn(self):

        return self.data_after_train
    
    def load_model_meta(self, model_meta):
        self.model_meta = model_meta
    
    def update_model_meta(self, dict):
        ''' update model meta
        '''
        self.model_meta.update(dict)

    def get_model_meta(self):
        return self.model_meta

    def save_model(self, fpath):
        ''' save model
        '''
        torch.save(self.model.state_dict(), fpath)

    def load_model(self, model_path = ''):
        ''' load model
        '''
        try:
            self.model.load_state_dict(torch.load(model_path, map_location=self.device))
        except:
            print(f"[BasePolicy.load_model] load model from {model_path} failed, please check the model path!")

class ToyPolicy:
    ''' base policy for traditional RL or non DRL
    '''
    def __init__(self, cfg) -> None:
        super().__init__()
        self.cfg = cfg
        self.obs_space = cfg.obs_space
        self.action_space = cfg.action_space
        self.get_state_action_size()
        self.policy_transition = {}
        self.data_after_train = {}
        self.Q_table = defaultdict(lambda: np.zeros(self.n_actions))

    def get_state_action_size(self):
        self.n_states = self.obs_space.n
        self.n_actions = self.action_space.n
    def create_summary(self):
        ''' create policy summary
        '''
        self.summary = {
            'scalar': {
                'loss': 0.0,
            },
        }
    def update_summary(self):
        ''' update policy summary
        '''
        self.summary['scalar']['loss'] = self.loss
        
    def get_summary(self):
        return self.summary['scalar']
    
    def get_action(self, state, **kwargs):
        ''' get action
        '''
        mode = kwargs.get('mode', None)
        if mode is None: 
            mode = 'sample' if self.cfg.mode == 'train' else 'predict'
        if mode == 'sample':
            return self.sample_action(state, **kwargs)
        elif mode == 'predict':
            return self.predict_action(state, **kwargs)
        else:
            raise NameError('mode must be sample or predict')
        
    def sample_action(self, state, **kwargs):
        raise NotImplementedError
    def predict_action(self, state, **kwargs):
        raise NotImplementedError
    def update_policy_transition(self):
        ''' update policy transition
        '''
        self.policy_transition = {}
    def get_policy_transition(self):
        return self.policy_transition
    def update_data_after_learn(self):
        ''' update data after training
        '''
        self.data_after_train = {}
    def learn(self, **kwargs):
        ''' learn policy
        '''
        raise NotImplementedError
    def get_model_params(self):
        ''' get model parameters
        '''
        return copy.deepcopy(self.Q_table)
    
    def put_model_params(self, params):
        ''' set model parameters
        '''
        self.Q_table = params
    def save_model(self, fpath):
        ''' save model
        '''
        torch.save(obj=self.Q_table, f=fpath, pickle_module=dill)
    def load_model(self, fpath):
        ''' load model
        '''
        self.Q_table = None
        self.Q_table = torch.load(fpath, pickle_module=dill)


#!/usr/bin/env python
# coding=utf-8
'''
Author: JiangJi
Email: johnjim0816@gmail.com
Date: 2024-05-26 23:54:07
LastEditor: JiangJi
LastEditTime: 2024-06-14 09:31:42
Discription: 
'''
import os
import math
import copy
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import random
import ray
from pathlib import Path
import matplotlib.pyplot as plt
from matplotlib import animation
import seaborn as sns
import yaml
import pandas as pd
from functools import wraps
from time import time
import logging
import ray
from threading import Lock
from queue import Queue, Empty
from matplotlib.font_manager import FontProperties  # 导入字体模块
from collections import deque
from pathlib import Path
@ray.remote
class DataActor:
    ''' 
    '''
    def __init__(self, initial_value: float | int | dict) -> None:
        self.value = initial_value
        self.lock = Lock()

    def get_value(self):
        with self.lock:
            return self.value

    def set_value(self, new_value):
        with self.lock:
            self.value = new_value

@ray.remote
class QueueActor:
    def __init__(self, maxsize: int = 0):
        self.queue = deque() if maxsize == 0 else deque(maxlen=maxsize)

    def put(self, item):
        self.queue.append(item)

    def pop(self):
        if len(self.queue) == 0:
            return None
        return self.queue.pop()
    
    def get(self):
        if len(self.queue) == 0:
            return None
        return self.queue.popleft()
    
class DeQueue(Queue):
    ''' Creating a thread-safe dequeue
    '''
    def pop(self, block=True, timeout=None):
        '''Remove and return an item from the queue.
        If optional args 'block' is true and 'timeout' is None (the default),
        block if necessary until an item is available. If 'timeout' is
        a non-negative number, it blocks at most 'timeout' seconds and raises
        the Empty exception if no item was available within that time.
        Otherwise ('block' is false), return an item if one is immediately
        available, else raise the Empty exception ('timeout' is ignored
        in that case).
        '''
        with self.not_empty:
            if not block:
                if not self._qsize():
                    raise Empty
            elif timeout is None:
                while not self._qsize():
                    self.not_empty.wait()
            elif timeout < 0:
                raise ValueError("'timeout' must be a non-negative number")
            else:
                endtime = time() + timeout
                while not self._qsize():
                    remaining = endtime - time()
                    if remaining <= 0.0:
                        raise Empty
                    self.not_empty.wait(remaining)
            item = self._pop()
            self.not_full.notify()
            return item
    
    def pop_nowait(self):
        '''Remove and return an item from the queue without blocking.
        Only get an item if one is immediately available. Otherwise
        raise the Empty exception.
        '''
        return self.pop(block=False)
    
    # Get an item from the queue right
    def _pop(self):
        return self.queue.pop()
    
    def append(self, item):
        ''' Put an item into the queue.
        '''
        with self.mutex:
            self.queue.append(item)
            self.unfinished_tasks += 1
class Logger(object):
    ''' Logger for print log to console
    '''
    def __init__(self, log_dir, *args, **kwargs) -> None:
        self.name = kwargs.get('log_name', '')
        self.logger = logging.getLogger(name = f"Log_{self.name}")
        self.logger.setLevel(logging.INFO) # default level is INFO
        self.formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s: - %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S')
        # output to file by using FileHandler
        if not self.logger.handlers: # avoid duplicate print
            name = self.name.split('_')[0]
            log_dir = f"{log_dir}/{name}"
            Path(log_dir).mkdir(parents=True, exist_ok=True)
            fh = logging.FileHandler(f"{log_dir}/Log_{self.name}.txt")
            fh.setLevel(logging.DEBUG)
            fh.setFormatter(self.formatter)
            self.logger.addHandler(fh)
            ch = logging.StreamHandler()
            ch.setLevel(logging.INFO)
            ch.setFormatter(self.formatter)
            self.logger.addHandler(ch)
    
    def info(self, content):
        ''' print info
        '''
        self.logger.info(content)
    
    def warning(self, content):
        ''' print warning
        '''
        self.logger.warning(content)

def load_model_meta(dir):
    ''' load model meta
    '''
    try:
        with open(f"{dir}/model_meta.yaml") as f:
            model_meta = yaml.load(f, Loader=yaml.FullLoader)
    except Exception as e:
        model_meta = {}
    return model_meta

def save_model_meta(dir, model_meta):
    ''' save model meta
    '''
    with open(f"{dir}/model_meta.yaml", 'w') as f:
        yaml.dump(model_meta, f, default_flow_style=False)
        
def chinese_font():
    ''' 设置中文字体，注意需要根据自己电脑情况更改字体路径，否则还是默认的字体
    '''
    try:
        font = FontProperties(
        fname='/System/Library/Fonts/STHeiti Light.ttc', size=15) # fname系统字体路径，此处是mac的
    except:
        font = None
    return font

def plot_rewards_cn(rewards, ma_rewards, cfg, tag='train'):
    ''' 中文画图
    '''
    sns.set()
    plt.figure()
    plt.title(u"{}环境下{}算法的学习曲线".format(cfg.env_name,
              cfg.algo_name), fontproperties=chinese_font())
    plt.xlabel(u'回合数', fontproperties=chinese_font())
    plt.plot(rewards)
    plt.plot(ma_rewards)
    plt.legend((u'奖励', u'滑动平均奖励',), loc="best", prop=chinese_font())
    if cfg.save:
        plt.savefig(cfg.result_path+f"{tag}_rewards_curve_cn")

def save_frames_as_gif(frames, fpath=None):
    ''' 保存gif动图
    '''
    Path(fpath).mkdir(parents=True, exist_ok=True)
    #Mess with this to change frame size
    plt.figure(figsize=(frames[0].shape[1] / 72.0, frames[0].shape[0] / 72.0), dpi=72)

    patch = plt.imshow(frames[0])
    plt.axis('off')

    def animate(i):
        patch.set_data(frames[i])

    anim = animation.FuncAnimation(plt.gcf(), animate, frames = len(frames), interval=50)
    anim.save(f"{fpath}/video.gif", writer='pillow', fps=60)
    
def smooth(data, weight=0.9):  
    '''用于平滑曲线，类似于Tensorboard中的smooth

    Args:
        data (List):输入数据
        weight (Float): 平滑权重，处于0-1之间，数值越高说明越平滑，一般取0.9

    Returns:
        smoothed (List): 平滑后的数据
    '''
    last = data[0]  # First value in the plot (first timestep)
    smoothed = list()
    for point in data:
        smoothed_val = last * weight + (1 - weight) * point  # 计算平滑值
        smoothed.append(smoothed_val)                    
        last = smoothed_val                                
    return smoothed

def plot_rewards(rewards,title="learning curve",fpath=None,save_fig=True,show_fig=False):
    sns.set()
    plt.figure()  # 创建一个图形实例，方便同时多画几个图
    plt.title(f"{title}")
    plt.xlabel('episodes')
    plt.plot(rewards, label='rewards')
    plt.plot(smooth(rewards), label='smoothed')
    plt.legend()
    if save_fig:
        plt.savefig(f"{fpath}/learning_curve.png")
    if show_fig:
        plt.show()

def plot_losses(losses, algo="DQN", save=True, path='./'):
    sns.set()
    plt.figure()
    plt.title("loss curve of {}".format(algo))
    plt.xlabel('episodes')
    plt.plot(losses, label='rewards')
    plt.legend()
    if save:
        plt.savefig(path+"losses_curve")
    plt.show()

def save_results(res_dic,fpath = None):
    ''' save results
    '''
    Path(fpath).mkdir(parents=True, exist_ok=True)
    df = pd.DataFrame(res_dic)
    df.to_csv(f"{fpath}/res.csv",index=None)
def merge_class_attrs(ob1, ob2):
    ob1.__dict__.update(ob2.__dict__)
    return ob1

def get_logger(fpath):
    Path(fpath).mkdir(parents=True, exist_ok=True)
    # logger = logging.getLogger(name='r')  # set root logger if not set name
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s: - %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S')
    # output to file by using FileHandler
    fh = logging.FileHandler(f"{fpath}/log.txt")
    fh.setLevel(logging.DEBUG)
    fh.setFormatter(formatter)
    # output to screen by using StreamHandler
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    ch.setFormatter(formatter)
    # add Handler
    logger.addHandler(ch)
    logger.addHandler(fh)
    return logger

def load_cfgs(cfgs, fpath):
    with open(fpath) as f:
        load_cfg = yaml.load(f, Loader=yaml.FullLoader)
        for cfg_type in cfgs:
            for k, v in load_cfg[cfg_type].items():
                setattr(cfgs[cfg_type], k, v)

def timing(func_name=True):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            start_time = time.time()
            result = func(*args, **kwargs)
            end_time = time.time()
            
            execution_time = end_time - start_time
            if func_name:
                print(f"func {func.__name__} executed: {execution_time} seconds")
            else:
                print(f"executed: {execution_time} seconds秒")
            return result
        return wrapper
    return decorator

def create_module(class_name, is_remote, remote_options, *args, **kwargs):
    ''' create module, which can be remote or local

    Args:
        remote (bool): remote or not
        options (dict): options for remote
    '''
    if is_remote:
        kwargs.update({'use_ray': True})
        return ray.remote(class_name).options(**remote_options).remote(*args, **kwargs)
    else:
        return class_name(*args, **kwargs)

    
def exec_method(instance_name, method_name, way: str, *args, **kwargs):
    ''' execute method of class_name, which can be remote or local

    Args:
        class_name (_type_): _description_
        method_name (_type_): _description_
        is_remote (bool, optional): remote or not. Defaults to False.
        need_return (bool, optional): need to return. Defaults to False.

    Returns:
        _type_: default None
    '''
    is_remote = isinstance(instance_name, ray.actor.ActorHandle)
    if is_remote:
        if way == "get":
            return ray.get(getattr(instance_name, method_name).remote(*args, **kwargs))
        elif way == "wait":
            return ray.wait([getattr(instance_name, method_name).remote(*args, **kwargs)])
        else:
            return getattr(instance_name, method_name).remote(*args, **kwargs)
    else:
        return getattr(instance_name, method_name)(*args, **kwargs)

def all_seed(seed = 1):
    ''' 设置随机种子，保证实验可复现，同时保证GPU和CPU的随机种子一致
    '''
    if seed == 0: # 值为0时不设置随机种子
        return
    os.environ['PYTHONHASHSEED'] = str(seed) # set PYTHONHASHSEED env var at fixed value
    np.random.seed(seed)
    random.seed(seed)
    torch.manual_seed(seed) # 
    torch.cuda.manual_seed(seed) # config for GPU
    # config for cudnn
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.enabled = False
    
def save_traj(traj, fpath):
    import pickle
    # create path
    Path(fpath).mkdir(parents=True, exist_ok=True)
    traj_pkl = os.path.join(fpath, 'traj.pkl')
    with open(traj_pkl, 'wb') as f:
        pickle.dump(traj, f)


        
# MAPPO beginning

# init
def init(module, weight_init, bias_init, gain=1):
    weight_init(module.weight.data, gain=gain)
    bias_init(module.bias.data)
    return module

def get_clones(module, N):
    return nn.ModuleList([copy.deepcopy(module) for i in range(N)])

def check(input):
    output = torch.from_numpy(input) if type(input) == np.ndarray else input
    return output

# CNN modules and utils
class Flatten(nn.Module):
    def forward(self, x):
        return x.view(x.size(0), -1)
class CNNLayer(nn.Module):
    def __init__(self, obs_shape, hidden_size, use_orthogonal, use_ReLU, kernel_size=3, stride=1):
        super(CNNLayer, self).__init__()

        active_func = [nn.Tanh(), nn.ReLU()][use_ReLU]
        init_method = [nn.init.xavier_uniform_, nn.init.orthogonal_][use_orthogonal]
        gain = nn.init.calculate_gain(['tanh', 'relu'][use_ReLU])

        def init_(m):
            return init(m, init_method, lambda x: nn.init.constant_(x, 0), gain=gain)

        input_channel = obs_shape[0]
        input_width = obs_shape[1]
        input_height = obs_shape[2]

        self.cnn = nn.Sequential(
            init_(nn.Conv2d(in_channels=input_channel,
                            out_channels=hidden_size // 2,
                            kernel_size=kernel_size,
                            stride=stride)
                  ),
            active_func,
            Flatten(),
            init_(nn.Linear(hidden_size // 2 * (input_width - kernel_size + stride) * (input_height - kernel_size + stride),
                            hidden_size)
                  ),
            active_func,
            init_(nn.Linear(hidden_size, hidden_size)), active_func)

    def forward(self, x):
        x = x / 255.0
        x = self.cnn(x)
        return x
class CNNBase(nn.Module):
    def __init__(self, args, obs_shape):
        super(CNNBase, self).__init__()

        self._use_orthogonal = args.use_orthogonal
        self._use_ReLU = args.use_ReLU
        self.hidden_size = args.hidden_size

        self.cnn = CNNLayer(obs_shape, self.hidden_size, self._use_orthogonal, self._use_ReLU)

    def forward(self, x):
        x = self.cnn(x)
        return x

# MLP modules
class MLPLayer(nn.Module):
    def __init__(self, input_dim, hidden_size, layer_N, use_orthogonal, use_ReLU):
        super(MLPLayer, self).__init__()
        self._layer_N = layer_N

        active_func = [nn.Tanh(), nn.ReLU()][use_ReLU]
        init_method = [nn.init.xavier_uniform_, nn.init.orthogonal_][use_orthogonal]
        gain = nn.init.calculate_gain(['tanh', 'relu'][use_ReLU])

        def init_(m):
            return init(m, init_method, lambda x: nn.init.constant_(x, 0), gain=gain)

        self.fc1 = nn.Sequential(
            init_(nn.Linear(input_dim, hidden_size)), active_func, nn.LayerNorm(hidden_size))
        self.fc_h = nn.Sequential(init_(
            nn.Linear(hidden_size, hidden_size)), active_func, nn.LayerNorm(hidden_size))
        self.fc2 = get_clones(self.fc_h, self._layer_N)

    def forward(self, x):
        x = self.fc1(x)
        for i in range(self._layer_N):
            x = self.fc2[i](x)
        return x

class MLPBase(nn.Module):
    def __init__(self, args, obs_shape, cat_self=True, attn_internal=False):
        super(MLPBase, self).__init__()

        self._use_feature_normalization = args.use_feature_normalization
        self._use_orthogonal = args.use_orthogonal
        self._use_ReLU = args.use_ReLU
        self._stacked_frames = args.stacked_frames
        self._layer_N = args.layer_N
        self.hidden_size = args.hidden_size

        obs_dim = obs_shape[0]

        if self._use_feature_normalization:
            self.feature_norm = nn.LayerNorm(obs_dim)

        self.mlp = MLPLayer(obs_dim, self.hidden_size,
                              self._layer_N, self._use_orthogonal, self._use_ReLU)

    def forward(self, x):
        if self._use_feature_normalization:
            x = self.feature_norm(x)

        x = self.mlp(x)

        return x

# RNN modules
class RNNLayer(nn.Module):
    def __init__(self, inputs_dim, outputs_dim, recurrent_N, use_orthogonal):
        super(RNNLayer, self).__init__()
        self._recurrent_N = recurrent_N
        self._use_orthogonal = use_orthogonal

        self.rnn = nn.GRU(inputs_dim, outputs_dim, num_layers=self._recurrent_N)
        for name, param in self.rnn.named_parameters():
            if 'bias' in name:
                nn.init.constant_(param, 0)
            elif 'weight' in name:
                if self._use_orthogonal:
                    nn.init.orthogonal_(param)
                else:
                    nn.init.xavier_uniform_(param)
        self.norm = nn.LayerNorm(outputs_dim)

    def forward(self, x, hxs, masks):
        if x.size(0) == hxs.size(0):
            x, hxs = self.rnn(x.unsqueeze(0),
                              (hxs * masks.repeat(1, self._recurrent_N).unsqueeze(-1)).transpose(0, 1).contiguous())
            x = x.squeeze(0)
            hxs = hxs.transpose(0, 1)
        else:
            # x is a (T, N, -1) tensor that has been flatten to (T * N, -1)
            N = hxs.size(0)
            T = int(x.size(0) / N)

            # unflatten
            x = x.view(T, N, x.size(1))

            # Same deal with masks
            masks = masks.view(T, N)

            # Let's figure out which steps in the sequence have a zero for any agent
            # We will always assume t=0 has a zero in it as that makes the logic cleaner
            has_zeros = ((masks[1:] == 0.0)
                         .any(dim=-1)
                         .nonzero()
                         .squeeze()
                         .cpu())

            # +1 to correct the masks[1:]
            if has_zeros.dim() == 0:
                # Deal with scalar
                has_zeros = [has_zeros.item() + 1]
            else:
                has_zeros = (has_zeros + 1).numpy().tolist()

            # add t=0 and t=T to the list
            has_zeros = [0] + has_zeros + [T]

            hxs = hxs.transpose(0, 1)

            outputs = []
            for i in range(len(has_zeros) - 1):
                # We can now process steps that don't have any zeros in masks together!
                # This is much faster
                start_idx = has_zeros[i]
                end_idx = has_zeros[i + 1]
                temp = (hxs * masks[start_idx].view(1, -1, 1).repeat(self._recurrent_N, 1, 1)).contiguous()
                rnn_scores, hxs = self.rnn(x[start_idx:end_idx], temp)
                outputs.append(rnn_scores)

            # assert len(outputs) == T
            # x is a (T, N, -1) tensor
            x = torch.cat(outputs, dim=0)

            # flatten
            x = x.reshape(T * N, -1)
            hxs = hxs.transpose(0, 1)

        x = self.norm(x)
        return x, hxs

# distribution modules
# Modify standard PyTorch distributions in order to make compatible with MAPPO codes
class FixedCategorical(torch.distributions.Categorical): # Categorical
    def sample(self):
        return super().sample().unsqueeze(-1)

    def log_probs(self, actions):
        return (
            super()
            .log_prob(actions.squeeze(-1))
            .view(actions.size(0), -1)
            .sum(-1)
            .unsqueeze(-1)
        )

    def mode(self):
        return self.probs.argmax(dim=-1, keepdim=True)

class FixedNormal(torch.distributions.Normal): # Normal
    def log_probs(self, actions):
        return super().log_prob(actions).sum(-1, keepdim=True)

    def entrop(self):
        return super.entropy().sum(-1)

    def mode(self):
        return self.mean

class FixedBernoulli(torch.distributions.Bernoulli): # Bernoulli
    def log_probs(self, actions):
        return super.log_prob(actions).view(actions.size(0), -1).sum(-1).unsqueeze(-1)

    def entropy(self):
        return super().entropy().sum(-1)

    def mode(self):
        return torch.gt(self.probs, 0.5).float()

class Categorical(nn.Module):
    def __init__(self, num_inputs, num_outputs, use_orthogonal=True, gain=0.01):
        super(Categorical, self).__init__()
        init_method = [nn.init.xavier_uniform_, nn.init.orthogonal_][use_orthogonal]
        def init_(m): 
            return init(m, init_method, lambda x: nn.init.constant_(x, 0), gain)

        self.linear = init_(nn.Linear(num_inputs, num_outputs))

    def forward(self, x, available_actions=None):
        x = self.linear(x)
        if available_actions is not None:
            x[available_actions == 0] = -1e10
        return FixedCategorical(logits=x)

class DiagGaussian(nn.Module):
    def __init__(self, num_inputs, num_outputs, use_orthogonal=True, gain=0.01):
        super(DiagGaussian, self).__init__()

        init_method = [nn.init.xavier_uniform_, nn.init.orthogonal_][use_orthogonal]
        def init_(m): 
            return init(m, init_method, lambda x: nn.init.constant_(x, 0), gain)

        self.fc_mean = init_(nn.Linear(num_inputs, num_outputs))
        self.logstd = AddBias(torch.zeros(num_outputs))

    def forward(self, x):
        action_mean = self.fc_mean(x)

        #  An ugly hack for my KFAC implementation.
        zeros = torch.zeros(action_mean.size())
        if x.is_cuda:
            zeros = zeros.cuda()
        
        action_logstd = self.logstd(zeros)
        return FixedNormal(action_mean, action_logstd.exp())

class Bernoulli(nn.Module):
    def __init__(self, num_inputs, num_outputs, use_orthogonal=True, gain=0.01):
        super(Bernoulli, self).__init__()
        init_method = [nn.init.xavier_uniform_, nn.init.orthogonal_][use_orthogonal]
        def init_(m): 
            return init(m, init_method, lambda x: nn.init.constant_(x, 0), gain)
        
        self.linear = init_(nn.Linear(num_inputs, num_outputs))

    def forward(self, x):
        x = self.linear(x)
        return FixedBernoulli(logits=x)

class AddBias(nn.Module):
    def __init__(self, bias):
        super(AddBias, self).__init__()
        self._bias = nn.Parameter(bias.unsqueeze(1))

    def forward(self, x):
        if x.dim() == 2:
            bias = self._bias.t().view(1, -1)
        else:
            bias = self._bias.t().view(1, -1, 1, 1)

        return x + bias

# using MLP modules to compute actions and probilities
class ACTLayer(nn.Module):
    """
    MLP Module to compute actions.
    :param action_space: (gym.Space) action space.
    :param inputs_dim: (int) dimension of network input.
    :param use_orthogonal: (bool) whether to use orthogonal initialization.
    :param gain: (float) gain of the output layer of the network.
    """
    def __init__(self, action_space, inputs_dim, use_orthogonal, gain):
        super(ACTLayer, self).__init__()
        self.mixed_action = False
        self.multi_discrete = False
        self.continuous_action = False

        if action_space.__class__.__name__ == "Discrete":
            action_dim = action_space.n
            self.action_out = Categorical(inputs_dim, action_dim, use_orthogonal, gain)
        elif action_space.__class__.__name__ == "Box":
            self.continuous_action = True
            action_dim = action_space.shape[0]
            self.action_out = DiagGaussian(inputs_dim, action_dim, use_orthogonal, gain)
        elif action_space.__class__.__name__ == "MultiBinary":
            action_dim = action_space.shape[0]
            self.action_out = Bernoulli(inputs_dim, action_dim, use_orthogonal, gain)
        elif action_space.__class__.__name__ == "MultiDiscrete":
            self.multi_discrete = True
            action_dims = action_space.high - action_space.low + 1
            self.action_outs = []
            for action_dim in action_dims:
                self.action_outs.append(Categorical(inputs_dim, action_dim, use_orthogonal, gain))
            self.action_outs = nn.ModuleList(self.action_outs)
        else:  # discrete + continous
            self.mixed_action = True
            continous_dim = action_space[0].shape[0]
            discrete_dim = action_space[1].n
            self.action_outs = nn.ModuleList([DiagGaussian(inputs_dim, continous_dim, use_orthogonal, gain), Categorical(
                inputs_dim, discrete_dim, use_orthogonal, gain)])
    
    def forward(self, x, available_actions=None, deterministic=False):
        """
        Compute actions and action logprobs from given input.
        :param x: (torch.Tensor) input to network.
        :param available_actions: (torch.Tensor) denotes which actions are available to agent
                                  (if None, all actions available)
        :param deterministic: (bool) whether to sample from action distribution or return the mode.

        :return actions: (torch.Tensor) actions to take.
        :return action_log_probs: (torch.Tensor) log probabilities of taken actions.
        """
        if self.mixed_action:
            actions = []
            action_log_probs = []
            for action_out in self.action_outs:
                action_logit = action_out(x)
                action = action_logit.mode() if deterministic else action_logit.sample()
                action_log_prob = action_logit.log_probs(action)
                actions.append(action.float())
                action_log_probs.append(action_log_prob)

            actions = torch.cat(actions, -1)
            action_log_probs = torch.sum(torch.cat(action_log_probs, -1), -1, keepdim=True)

        elif self.multi_discrete:
            actions = []
            action_log_probs = []
            for action_out in self.action_outs:
                action_logit = action_out(x)
                action = action_logit.mode() if deterministic else action_logit.sample()
                action_log_prob = action_logit.log_probs(action)
                actions.append(action)
                action_log_probs.append(action_log_prob)

            actions = torch.cat(actions, -1)
            action_log_probs = torch.cat(action_log_probs, -1)
        elif self.continuous_action:
            # actions = []
            # action_log_probs = []
            action_logit = self.action_out(x)
            actions = action_logit.mode() if deterministic else action_logit.sample()
            action_log_probs = action_logit.log_probs(actions)
            # actions.append(action.float())
            # action_log_probs.append(action_log_prob)
            # actions = torch.cat(actions, -1)
            # action_log_probs = torch.sum(torch.cat(action_log_probs, -1), -1, keepdim=True)
        else:
            action_logits = self.action_out(x, available_actions)
            actions = action_logits.mode() if deterministic else action_logits.sample() 
            action_log_probs = action_logits.log_probs(actions)
        
        return actions, action_log_probs

    def get_probs(self, x, available_actions=None):
        """
        Compute action probabilities from inputs.
        :param x: (torch.Tensor) input to network.
        :param available_actions: (torch.Tensor) denotes which actions are available to agent
                                  (if None, all actions available)

        :return action_probs: (torch.Tensor)
        """
        if self.mixed_action or self.multi_discrete:
            action_probs = []
            for action_out in self.action_outs:
                action_logit = action_out(x)
                action_prob = action_logit.probs
                action_probs.append(action_prob)
            action_probs = torch.cat(action_probs, -1)
        else:
            action_logits = self.action_out(x, available_actions)
            action_probs = action_logits.probs
        
        return action_probs

    def evaluate_actions(self, x, action, available_actions=None, active_masks=None):
        """
        Compute log probability and entropy of given actions.
        :param x: (torch.Tensor) input to network.
        :param action: (torch.Tensor) actions whose entropy and log probability to evaluate.
        :param available_actions: (torch.Tensor) denotes which actions are available to agent
                                                              (if None, all actions available)
        :param active_masks: (torch.Tensor) denotes whether an agent is active or dead.

        :return action_log_probs: (torch.Tensor) log probabilities of the input actions.
        :return dist_entropy: (torch.Tensor) action distribution entropy for the given inputs.
        """
        if self.mixed_action:
            a, b = action.split((2, 1), -1)
            b = b.long()
            action = [a, b] 
            action_log_probs = [] 
            dist_entropy = []
            for action_out, act in zip(self.action_outs, action):
                action_logit = action_out(x)
                action_log_probs.append(action_logit.log_probs(act))
                if active_masks is not None:
                    if len(action_logit.entropy().shape) == len(active_masks.shape):
                        dist_entropy.append((action_logit.entropy() * active_masks).sum()/active_masks.sum()) 
                    else:
                        dist_entropy.append((action_logit.entropy() * active_masks.squeeze(-1)).sum()/active_masks.sum())
                else:
                    dist_entropy.append(action_logit.entropy().mean())
                
            action_log_probs = torch.sum(torch.cat(action_log_probs, -1), -1, keepdim=True)
            dist_entropy = dist_entropy[0] / 2.0 + dist_entropy[1] / 0.98 #! dosen't make sense

        elif self.multi_discrete:
            action = torch.transpose(action, 0, 1)
            action_log_probs = []
            dist_entropy = []
            for action_out, act in zip(self.action_outs, action):
                action_logit = action_out(x)
                action_log_probs.append(action_logit.log_probs(act))
                if active_masks is not None:
                    dist_entropy.append((action_logit.entropy()*active_masks.squeeze(-1)).sum()/active_masks.sum())
                else:
                    dist_entropy.append(action_logit.entropy().mean())

            action_log_probs = torch.cat(action_log_probs, -1) # ! could be wrong
            dist_entropy = torch.tensor(dist_entropy).mean()

        elif self.continuous_action:
            # a, b = action.split((2, 1), -1)
            # b = b.long()
            # action = [a, b]
            action_log_probs = []
            dist_entropy = []
            # for action_out, act in zip(self.action_outs, action):
            action_logit = self.action_out(x)
            action_log_probs.append(action_logit.log_probs(action))
            if active_masks is not None:
                if len(action_logit.entropy().shape) == len(active_masks.shape):
                    dist_entropy.append((action_logit.entropy() * active_masks).sum() / active_masks.sum())
                else:
                    dist_entropy.append(
                        (action_logit.entropy() * active_masks.squeeze(-1)).sum() / active_masks.sum())
            else:
                dist_entropy.append(action_logit.entropy().mean())

            action_log_probs = torch.sum(torch.cat(action_log_probs, -1), -1, keepdim=True)
            dist_entropy = dist_entropy[0] # / 2.0 + dist_entropy[1] / 0.98  # ! dosen't make sense

        else:
            action_logits = self.action_out(x, available_actions)
            action_log_probs = action_logits.log_probs(action)
            if active_masks is not None:
                dist_entropy = (action_logits.entropy()*active_masks.squeeze(-1)).sum()/active_masks.sum()
            else:
                dist_entropy = action_logits.entropy().mean()
        
        return action_log_probs, dist_entropy


# popart
class PopArt(torch.nn.Module):
    
    def __init__(self, input_shape, output_shape, norm_axes=1, beta=0.99999, epsilon=1e-5, device=torch.device("cpu")):
        
        super(PopArt, self).__init__()

        self.beta = beta
        self.epsilon = epsilon
        self.norm_axes = norm_axes
        self.tpdv = dict(dtype=torch.float32, device=device)

        self.input_shape = input_shape
        self.output_shape = output_shape

        self.weight = nn.Parameter(torch.Tensor(output_shape, input_shape)).to(**self.tpdv)
        self.bias = nn.Parameter(torch.Tensor(output_shape)).to(**self.tpdv)
        
        self.stddev = nn.Parameter(torch.ones(output_shape), requires_grad=False).to(**self.tpdv)
        self.mean = nn.Parameter(torch.zeros(output_shape), requires_grad=False).to(**self.tpdv)
        self.mean_sq = nn.Parameter(torch.zeros(output_shape), requires_grad=False).to(**self.tpdv)
        self.debiasing_term = nn.Parameter(torch.tensor(0.0), requires_grad=False).to(**self.tpdv)

        self.reset_parameters()

    def reset_parameters(self):
        torch.nn.init.kaiming_uniform_(self.weight, a=math.sqrt(5))
        if self.bias is not None:
            fan_in, _ = torch.nn.init._calculate_fan_in_and_fan_out(self.weight)
            bound = 1 / math.sqrt(fan_in)
            torch.nn.init.uniform_(self.bias, -bound, bound)
        self.mean.zero_()
        self.mean_sq.zero_()
        self.debiasing_term.zero_()

    def forward(self, input_vector):
        if type(input_vector) == np.ndarray:
            input_vector = torch.from_numpy(input_vector)
        input_vector = input_vector.to(**self.tpdv)

        return F.linear(input_vector, self.weight, self.bias)
    
    @torch.no_grad()
    def update(self, input_vector):
        if type(input_vector) == np.ndarray:
            input_vector = torch.from_numpy(input_vector)
        input_vector = input_vector.to(**self.tpdv)
        
        old_mean, old_stddev = self.mean, self.stddev

        batch_mean = input_vector.mean(dim=tuple(range(self.norm_axes)))
        batch_sq_mean = (input_vector ** 2).mean(dim=tuple(range(self.norm_axes)))

        self.mean.mul_(self.beta).add_(batch_mean * (1.0 - self.beta))
        self.mean_sq.mul_(self.beta).add_(batch_sq_mean * (1.0 - self.beta))
        self.debiasing_term.mul_(self.beta).add_(1.0 * (1.0 - self.beta))

        self.stddev = (self.mean_sq - self.mean ** 2).sqrt().clamp(min=1e-4)

        self.weight = self.weight * old_stddev / self.stddev
        self.bias = (old_stddev * self.bias + old_mean - self.mean) / self.stddev

    def debiased_mean_var(self):
        debiased_mean = self.mean / self.debiasing_term.clamp(min=self.epsilon)
        debiased_mean_sq = self.mean_sq / self.debiasing_term.clamp(min=self.epsilon)
        debiased_var = (debiased_mean_sq - debiased_mean ** 2).clamp(min=1e-2)
        return debiased_mean, debiased_var

    def normalize(self, input_vector):
        if type(input_vector) == np.ndarray:
            input_vector = torch.from_numpy(input_vector)
        input_vector = input_vector.to(**self.tpdv)

        mean, var = self.debiased_mean_var()
        out = (input_vector - mean[(None,) * self.norm_axes]) / torch.sqrt(var)[(None,) * self.norm_axes]
        
        return out

    def denormalize(self, input_vector):
        if type(input_vector) == np.ndarray:
            input_vector = torch.from_numpy(input_vector)
        input_vector = input_vector.to(**self.tpdv)

        mean, var = self.debiased_mean_var()
        out = input_vector * torch.sqrt(var)[(None,) * self.norm_axes] + mean[(None,) * self.norm_axes]
        
        out = out.cpu().numpy()

        return out

# util
# def check(input):
#     if type(input) == np.ndarray:
#         return torch.from_numpy(input)
        
def get_gard_norm(it):
    sum_grad = 0
    for x in it:
        if x.grad is None:
            continue
        sum_grad += x.grad.norm() ** 2
    return math.sqrt(sum_grad)

def update_linear_schedule(optimizer, epoch, total_num_epochs, initial_lr):
    """Decreases the learning rate linearly"""
    lr = initial_lr - (initial_lr * (epoch / float(total_num_epochs)))
    for param_group in optimizer.param_groups:
        param_group['lr'] = lr

def huber_loss(e, d):
    a = (abs(e) <= d).float()
    b = (e > d).float()
    return a*e**2/2 + b*d*(abs(e)-d/2)

def mse_loss(e):
    return e**2/2

def get_shape_from_obs_space(obs_space):
    if obs_space.__class__.__name__ == 'Box':
        obs_shape = obs_space.shape
    elif obs_space.__class__.__name__ == 'list':
        obs_shape = obs_space
    else:
        raise NotImplementedError
    return obs_shape

def get_shape_from_act_space(act_space):
    if act_space.__class__.__name__ == 'Discrete':
        act_shape = 1
    elif act_space.__class__.__name__ == "MultiDiscrete":
        act_shape = act_space.shape
    elif act_space.__class__.__name__ == "Box":
        act_shape = act_space.shape[0]
    elif act_space.__class__.__name__ == "MultiBinary":
        act_shape = act_space.shape[0]
    else:  # agar
        act_shape = act_space[0].shape[0] + 1  
    return act_shape


def tile_images(img_nhwc):
    """
    Tile N images into one big PxQ image
    (P,Q) are chosen to be as close as possible, and if N
    is square, then P=Q.
    input: img_nhwc, list or array of images, ndim=4 once turned into array
        n = batch index, h = height, w = width, c = channel
    returns:
        bigim_HWc, ndarray with ndim=3
    """
    img_nhwc = np.asarray(img_nhwc)
    N, h, w, c = img_nhwc.shape
    H = int(np.ceil(np.sqrt(N)))
    W = int(np.ceil(float(N)/H))
    img_nhwc = np.array(list(img_nhwc) + [img_nhwc[0]*0 for _ in range(N, H*W)])
    img_HWhwc = img_nhwc.reshape(H, W, h, w, c)
    img_HhWwc = img_HWhwc.transpose(0, 2, 1, 3, 4)
    img_Hh_Ww_c = img_HhWwc.reshape(H*h, W*w, c)
    return img_Hh_Ww_c

# value normlization
class ValueNorm(nn.Module):
    """ Normalize a vector of observations - across the first norm_axes dimensions"""

    def __init__(self, input_shape, norm_axes=1, beta=0.99999, per_element_update=False, epsilon=1e-5, device=torch.device("cpu")):
        super(ValueNorm, self).__init__()

        self.input_shape = input_shape
        self.norm_axes = norm_axes
        self.epsilon = epsilon
        self.beta = beta
        self.per_element_update = per_element_update
        self.tpdv = dict(dtype=torch.float32, device=device)

        self.running_mean = nn.Parameter(torch.zeros(input_shape), requires_grad=False).to(**self.tpdv)
        self.running_mean_sq = nn.Parameter(torch.zeros(input_shape), requires_grad=False).to(**self.tpdv)
        self.debiasing_term = nn.Parameter(torch.tensor(0.0), requires_grad=False).to(**self.tpdv)
        
        self.reset_parameters()

    def reset_parameters(self):
        self.running_mean.zero_()
        self.running_mean_sq.zero_()
        self.debiasing_term.zero_()

    def running_mean_var(self):
        debiased_mean = self.running_mean / self.debiasing_term.clamp(min=self.epsilon)
        debiased_mean_sq = self.running_mean_sq / self.debiasing_term.clamp(min=self.epsilon)
        debiased_var = (debiased_mean_sq - debiased_mean ** 2).clamp(min=1e-2)
        return debiased_mean, debiased_var

    @torch.no_grad()
    def update(self, input_vector):
        if type(input_vector) == np.ndarray:
            input_vector = torch.from_numpy(input_vector)
        input_vector = input_vector.to(**self.tpdv)

        batch_mean = input_vector.mean(dim=tuple(range(self.norm_axes)))
        batch_sq_mean = (input_vector ** 2).mean(dim=tuple(range(self.norm_axes)))

        if self.per_element_update:
            batch_size = np.prod(input_vector.size()[:self.norm_axes])
            weight = self.beta ** batch_size
        else:
            weight = self.beta

        self.running_mean.mul_(weight).add_(batch_mean * (1.0 - weight))
        self.running_mean_sq.mul_(weight).add_(batch_sq_mean * (1.0 - weight))
        self.debiasing_term.mul_(weight).add_(1.0 * (1.0 - weight))

    def normalize(self, input_vector):
        # Make sure input is float32
        if type(input_vector) == np.ndarray:
            input_vector = torch.from_numpy(input_vector)
        input_vector = input_vector.to(**self.tpdv)

        mean, var = self.running_mean_var()
        out = (input_vector - mean[(None,) * self.norm_axes]) / torch.sqrt(var)[(None,) * self.norm_axes]
        
        return out

    def denormalize(self, input_vector):
        """ Transform normalized data back into original distribution """
        if type(input_vector) == np.ndarray:
            input_vector = torch.from_numpy(input_vector)
        input_vector = input_vector.to(**self.tpdv)

        mean, var = self.running_mean_var()
        out = input_vector * torch.sqrt(var)[(None,) * self.norm_axes] + mean[(None,) * self.norm_axes]
        
        out = out.cpu().numpy()
        
        return out
# MAPPO ending



#!/usr/bin/env python
# coding=utf-8
'''
Author: JiangJi
Email: johnjim0816@gmail.com
Date: 2023-12-15 13:16:24
LastEditor: JiangJi
LastEditTime: 2024-01-25 12:01:27
Discription: 
'''
import numpy as np
class AlgoConfig:
    def __init__(self):
        self.action_type_list = 'dpg' # action type, dpg: deterministic policy gradient
        self.buffer_type = 'REPLAY_QUE' # replay buffer type
        self.max_buffer_size = 100000  # replay buffer size
        self.batch_size = 128  # batch size
        self.gamma = 0.99  # discount factor
        self.policy_loss_weight = 0.002  # policy loss weight
        self.critic_lr = 1e-3  # learning rate of critic
        self.actor_lr = 1e-4  # learning rate of actor
        self.tau = 0.001  # soft update parameter
        self.value_min = -np.inf  # clip min critic value
        self.value_max = np.inf  # clip max critic value
        # self.actor_layers = [
        #     {'layer_type': 'Linear', 'layer_size': [256], 'activation': 'ReLU'},
        #     {'layer_type': 'Linear', 'layer_size': [256], 'activation': 'ReLU'},
        # ]
        # self.critic_layers = [
        #     {'layer_type': 'Linear', 'layer_size': [256], 'activation': 'ReLU'},
        #     {'layer_type': 'Linear', 'layer_size': [256], 'activation': 'ReLU'},
        # ]
        self.branch_layers = []
        self.merge_layers = []
        self.actor_branch_layers = []
        self.actor_merge_layers = []
        self.critic_branch_layers = []
        self.critic_merge_layers = []


#!/usr/bin/env python
# coding=utf-8
'''
Author: JiangJi
Email: johnjim0816@gmail.com
Date: 2023-12-22 23:02:13
LastEditor: JiangJi
LastEditTime: 2024-06-12 00:06:25
Discription: 
'''
import copy
import torch
import torch.nn as nn
from joyrl.algos.base.base_layer import create_layer, LayerConfig
from joyrl.algos.base.action_layer import ActionLayerType, DiscreteActionLayer, ContinuousActionLayer, DPGActionLayer, DQNActionLayer
from joyrl.framework.config import MergedConfig

class BranchLayers(nn.Module):
    def __init__(self, branch_layers_cfg : list, input_size_list, **kwargs) -> None:
        super(BranchLayers, self).__init__(**kwargs)
        self.input_size_list = input_size_list
        self.branch_layers = nn.ModuleList([nn.ModuleList() for _ in range(len(self.input_size_list))])
        self.output_size_list = copy.deepcopy(input_size_list)
        self.branch_layers_cfg = branch_layers_cfg
        if self.branch_layers_cfg is None:
            self.branch_layers_cfg = []
        for i, branch_layer_cfg in enumerate(self.branch_layers_cfg):
            if i >= len(input_size_list): # if branch_layers_cfg is more than input_size
                break
            input_size = input_size_list[i]
            layers_cfg_list = branch_layer_cfg['layers']
            for layer_cfg_dic in layers_cfg_list:
                if "layer_type" not in layer_cfg_dic:
                    raise ValueError("layer_type must be specified in layer_cfg")
                layer_cfg = LayerConfig(**layer_cfg_dic)
                layer, layer_output_size = create_layer(input_size, layer_cfg)
                self.branch_layers[i].append(layer)
                input_size = layer_output_size
            self.output_size_list[i] = layer_output_size
            
    def forward(self, x):
        if isinstance(x, torch.Tensor): # if x is a tensor, convert it to a list
            x = [x]
        for i, branch_layer in enumerate(self.branch_layers):
            for layer in branch_layer:
                x[i] = layer(x[i])
        return x
    
    def reset_noise(self):
        ''' reset noise for noisy layers
        '''
        for branch_layer in self.branch_layers:
            for layer in branch_layer:
                if hasattr(layer, "reset_noise"):
                    layer.reset_noise()
    
class MergeLayer(nn.Module):
    def __init__(self, merge_layers_cfg: list, input_size_list: list, **kwargs) -> None:
        super(MergeLayer, self).__init__(**kwargs)
        input_dim = sum([input_size[1] for input_size in input_size_list])
        self.input_size = [None, input_dim]
        self.output_size = copy.deepcopy(self.input_size)
        self.merge_layers = nn.ModuleList()
        self.merge_layers_cfg = merge_layers_cfg
        if self.merge_layers_cfg is None:
            self.merge_layers_cfg = []
        for layer_cfg_dic in self.merge_layers_cfg:
            if "layer_type" not in layer_cfg_dic:
                raise ValueError("layer_type must be specified in layer_cfg")
            layer_cfg = LayerConfig(**layer_cfg_dic)
            layer, layer_output_size = create_layer(self.input_size, layer_cfg)
            self.merge_layers.append(layer)
            self.input_size = layer_output_size
            self.output_size = layer_output_size

    def forward(self, x):
        x = torch.cat(x, dim=1)
        for layer in self.merge_layers:
            x = layer(x)
        return x
    def reset_noise(self):
        ''' reset noise for noisy layers
        '''
        for layer in self.merge_layers:
            if hasattr(layer, "reset_noise"):
                layer.reset_noise()

class ActionLayers(nn.Module):
    def __init__(self, cfg: MergedConfig, input_size: list) -> None:
        super(ActionLayers, self).__init__()
        self.cfg = cfg
        self.input_size = input_size
        self.action_size_list = self.cfg.action_size_list
        self.action_type_list = self.cfg.action_type_list
        assert len(self.action_type_list) == 1 and len(self.action_type_list) == len(self.action_size_list), "action_type_list and action_size_list must have the same length, and only one output"
        self.actor_outputs = []
        self.create_graph()

    def create_graph(self):
        self.action_layers = nn.ModuleList()
        for i in range(len(self.action_type_list)):
            action_type = ActionLayerType[self.action_type_list[i].upper()]
            action_size = self.action_size_list[i]
            if action_type == ActionLayerType.CONTINUOUS:
                action_layer = ContinuousActionLayer(self.cfg, self.input_size, action_size, id = i)
            elif action_type == ActionLayerType.DISCRETE:
                action_layer = DiscreteActionLayer(self.cfg, self.input_size, action_size,id = i)
            elif action_type == ActionLayerType.DPG:
                action_layer = DPGActionLayer(self.cfg, self.input_size, action_size, id = i)
            elif action_type == ActionLayerType.DQNACTION:
                action_layer = DQNActionLayer(self.cfg, self.input_size, action_size, id = i)
            else:
                raise ValueError("action_type must be specified in discrete, continuous or dpg")
            self.action_layers.append(action_layer)

    def forward(self, x, **kwargs):
        self.actor_outputs = {}
        for i, action_layer in enumerate(self.action_layers):
            self.actor_outputs[i] = action_layer(x, **kwargs)
        return self.actor_outputs
    
    def get_mus(self):
        mus = []
        for _, action_layer in enumerate(self.action_layers):
            mus.append(action_layer.get_mu())
        return mus[0]
    
    def get_qvalues(self):
        qvalues = []
        for _, action_layer in enumerate(self.action_layers):
            qvalues.append(action_layer.get_qvalue())
        return qvalues
    
    def get_actions(self, **kwargs):
        mode = kwargs.get('mode', 'train')
        actor_outputs = kwargs.get('actor_outputs', None)
        actions = []
        for i, action_layer in enumerate(self.action_layers):
            action_layer_output = action_layer.get_action(mode = mode, actor_output = actor_outputs[i])
            actions.append(action_layer_output['action'])
        return actions
    
    def get_actions_and_log_probs(self, **kwargs):
        mode = kwargs.get('mode', 'train')
        actor_outputs = kwargs.get('actor_outputs')
        actions = []
        log_probs_sum = 0
        for i, action_layer in enumerate(self.action_layers):
            action_layer_output = action_layer.get_action(mode = mode, actor_output = actor_outputs[i])
            actions.append(action_layer_output['action'])
            log_probs_sum += action_layer_output['log_prob']
        return actions, log_probs_sum
    
    def get_log_probs_action(self, actor_outputs, actions):
        log_prob_sum = 0
        for i, action_layer in enumerate(self.action_layers):
            actor_output = actor_outputs[i]
            action = actions[i]
            log_prob = action_layer.get_log_prob_action(actor_output, action)
            log_prob_sum += log_prob
        return log_prob_sum
    
    def get_mean_entropy(self, actor_outputs):
        entropy_sum = 0
        for i, action_layer in enumerate(self.action_layers):
            entropy_sum += action_layer.get_entropy(actor_outputs[i])
        entropy_mean = entropy_sum / len(self.action_layers)
        return entropy_mean
     
class BaseNework(nn.Module):
    def __init__(self, cfg: MergedConfig, input_size_list) -> None:
        super(BaseNework, self).__init__()
        self.cfg = cfg
        self.input_size_list = input_size_list
        
    def create_graph(self):
        self.branch_layers = BranchLayers(self.cfg.branch_layers, self.input_size_list)
        self.merge_layer = MergeLayer(self.cfg.merge_layers, self.branch_layers.output_size_list)

    def forward(self, x):
        x = self.branch_layers(x)
        x = self.merge_layer(x)
        return x
    
class QNetwork(BaseNework):
    ''' Q network, for value-based methods like DQN
    '''
    def __init__(self, cfg: MergedConfig, input_size_list: list) -> None:
        '''_summary_
        Args:
            cfg (_type_): _description_
            state_size (_type_): [[None, state_dim_1], [None, None, state_dim_2], ...]
            action_size (_type_): [action_dim_1, action_dim_2, ...]
        Raises:
            ValueError: _description_
        '''
        super(QNetwork, self).__init__(cfg, input_size_list)
        self.action_size_list = self.cfg.action_size_list
        self.dueling = hasattr(cfg, 'dueling') and cfg.dueling
        self.create_graph()

    def create_graph(self):
        self.branch_layers = BranchLayers(self.cfg.branch_layers, self.input_size_list)
        self.merge_layer = MergeLayer(self.cfg.merge_layers, self.branch_layers.output_size_list)
        action_type_list = ['dqnaction'] * len(self.cfg.action_type_list) 
        setattr(self.cfg, 'action_type_list', action_type_list)
        self.action_layers = ActionLayers(self.cfg, self.merge_layer.output_size)

    def forward(self, x):
        x = self.branch_layers(x)
        x = self.merge_layer(x)
        actor_outputs = self.action_layers(x)
        return {"actor_outputs": actor_outputs}
    
    def reset_noise(self):
        ''' reset noise for noisy layers
        '''
        self.branch_layers.reset_noise()
        self.merge_layer.reset_noise()


class ActorNetwork(BaseNework):
    def __init__(self, cfg: MergedConfig, input_size_list) -> None:
        super(ActorNetwork, self).__init__(cfg, input_size_list)
        self.action_type_list = self.cfg.action_type_list
        self.create_graph()
    
    def create_graph(self):
        self.branch_layers = BranchLayers(self.cfg.actor_branch_layers, self.input_size_list)
        self.merge_layer = MergeLayer(self.cfg.actor_merge_layers, self.branch_layers.output_size_list)
        self.action_layers = ActionLayers(self.cfg, self.merge_layer.output_size)
        
    def forward(self, x, pre_legal_actions=None):
        x = self.branch_layers(x)
        x = self.merge_layer(x)
        action_outputs = self.action_layers(x, pre_legal_actions = pre_legal_actions)
        return action_outputs

class CriticNetwork(BaseNework):
    def __init__(self, cfg: MergedConfig, input_size_list) -> None:
        super(CriticNetwork, self).__init__(cfg, input_size_list)
        self.create_graph()

    def create_graph(self):
        self.branch_layers = BranchLayers(self.cfg.critic_branch_layers, self.input_size_list)
        self.merge_layer = MergeLayer(self.cfg.critic_merge_layers, self.branch_layers.output_size_list)
        self.value_layer_cfg = LayerConfig(layer_type='linear', layer_size=[1], activation='none')
        self.value_layer, _ = create_layer(self.merge_layer.output_size, self.value_layer_cfg)
    
    def forward(self, x):
        x = self.branch_layers(x)
        x = self.merge_layer(x)
        value = self.value_layer(x)
        return value
    
class ActorCriticNetwork(BaseNework):
    ''' Value network, for policy-based methods,  in which the branch_layers and critic share the same network
    '''
    def __init__(self, cfg: MergedConfig, input_size_list: list) -> None:
        super(ActorCriticNetwork, self).__init__(cfg, input_size_list)
        self.action_type_list = self.cfg.action_type_list
        self.create_graph()
        
    def create_graph(self):
        if getattr(self.cfg, 'independ_actor', False):
            self.actor = ActorNetwork(self.cfg, self.input_size_list)
            self.critic = CriticNetwork(self.cfg, self.input_size_list)
        else:
            self.branch_layers = BranchLayers(self.cfg.branch_layers, self.input_size_list)
            self.merge_layer = MergeLayer(self.cfg.merge_layers, self.branch_layers.output_size_list)
            self.value_layer, _ = create_layer(self.merge_layer.output_size, LayerConfig(layer_type='linear', layer_size=[1], activation='none'))
            self.action_layers = ActionLayers(self.cfg, self.merge_layer.output_size,)
        
    def forward(self, x, pre_legal_actions=None):
        if getattr(self.cfg, 'independ_actor', False):
            # since input x is a list, need to deepcopy it to avoid changing the original x
            actor_outputs = self.actor(copy.deepcopy(x), pre_legal_actions)
            value = self.critic(copy.deepcopy(x))
            return {'value': value, 'actor_outputs': actor_outputs}
        else:
            x = self.branch_layers(x)
            x = self.merge_layer(x)
            value = self.value_layer(x)
            actor_outputs = self.action_layers(x, pre_legal_actions = pre_legal_actions)
            return {'value': value, 'actor_outputs': actor_outputs}
        
    def get_actions_and_log_probs(self, **kwargs):
        if getattr(self.cfg, 'independ_actor', False):
            return self.actor.action_layers.get_actions_and_log_probs(**kwargs)
        else:
            return self.action_layers.get_actions_and_log_probs(**kwargs)
        
    def get_log_probs_action(self, actor_outputs, actions):
        if getattr(self.cfg, 'independ_actor', False):
            return self.actor.action_layers.get_log_probs_action(actor_outputs, actions)
        else:
            return self.action_layers.get_log_probs_action(actor_outputs, actions)
        
    def get_mean_entropy(self, actor_outputs):
        if getattr(self.cfg, 'independ_actor', False):
            return self.actor.action_layers.get_mean_entropy(actor_outputs)
        else:
            return self.action_layers.get_mean_entropy(actor_outputs)
    
    def get_actions(self, **kwargs):
        if getattr(self.cfg, 'independ_actor', False):
            return self.actor.action_layers.get_actions(**kwargs)
        else:
            return self.action_layers.get_actions(**kwargs)
        



class AlgoConfig():
    ''' algorithm parameters
    '''
    def __init__(self) -> None:
        # set epsilon_start=epsilon_end to get fixed epsilon, i.e. epsilon=epsilon_end
        self.epsilon_start = 0.95  # epsilon start value
        self.epsilon_end = 0.01  # epsilon end value
        self.epsilon_decay = 500  # epsilon decay
        self.gamma = 0.95  # reward discount factor
        self.lr = 0.0001  # learning rate



import numpy as np

class OUNoise(object):
    ''' Ornstein–Uhlenbeck Noise
    '''
    def __init__(self, action_space, mu=0.0, theta=0.15, max_sigma=0.3, min_sigma=0.3, decay_period=10000):
        self.mu = mu
        self.theta = theta
        self.sigma = max_sigma
        self.max_sigma = max_sigma
        self.min_sigma = min_sigma
        self.decay_period = decay_period
        self.n_actions = action_space.shape[0] 
        self.low = action_space.low  
        self.high = action_space.high 
        self.reset()

    def reset(self):
        ''' reset the Ornstein–Uhlenbeck Noise
        '''
        self.obs = np.ones(self.n_actions) * self.mu  # reset the noise

    def evolve_obs(self):
        ''' evolove the Ornstein–Uhlenbeck Noise
        '''
        x = self.obs
        dx = self.theta * (self.mu - x) + self.sigma * np.random.randn(self.n_actions)  # Ornstein–Uhlenbeck process
        self.obs = x + dx
        return self.obs

    def get_action(self, action, t=0):
        ''' add noise to action
        '''
        ou_obs = self.evolve_obs()
        #  decay the action noise, as described in the paper
        self.sigma = self.max_sigma - (self.max_sigma - self.min_sigma) * min(1.0, t / self.decay_period)
        return np.clip(action + ou_obs, self.low, self.high)  


#!/usr/bin/env python
# coding=utf-8
'''
Author: JiangJi
Email: johnjim0816@gmail.com
Date: 2023-01-01 16:20:49
LastEditor: JiangJi
LastEditTime: 2023-12-25 12:52:47
Discription: 
'''
from joyrl.algos import base,DQN,DoubleDQN,DuelingDQN,NoisyDQN,PPO
__all__ = [
    "base",
    "QLearning",
    "DQN",
    "DoubleDQN",
    "DuelingDQN",
    "NoisyDQN",
    "PPO"
]
